# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Module for downloading and curating data from ECMWFs ERA5 dataset (via CDS).

For further reference see
https://confluence.ecmwf.int/display/CKB/ERA5%3A+data+documentation
"""

import logging
import os
import warnings
import weakref
from tempfile import mkstemp

import cdsapi
import numpy as np
import pandas as pd
import xarray as xr
from dask import compute, delayed
from dask.array import arctan2, sqrt
from numpy import atleast_1d

from atlite.gis import maybe_swap_spatial_dims
from atlite.pv.solar_position import SolarPosition

# Null context for running a with statements wihout any context
try:
    from contextlib import nullcontext
except ImportError:
    # for Python verions < 3.7:
    import contextlib

    @contextlib.contextmanager
    def nullcontext():
        yield


logger = logging.getLogger(__name__)

# Model and CRS Settings
crs = 4326

features = {
    "height": ["height"],
    "wind": ["wnd100m", "wnd_shear_exp", "wnd_azimuth", "roughness"],
    "influx": [
        "influx_toa",
        "influx_direct",
        "influx_diffuse",
        "albedo",
        "solar_altitude",
        "solar_azimuth",
    ],
    "temperature": ["temperature", "soil temperature", "dewpoint temperature"],
    "runoff": ["runoff"],
}

static_features = {"height"}


def _add_height(ds):
    """
    Convert geopotential 'z' to geopotential height following [1].

    References
    ----------
    [1] ERA5: surface elevation and orography, retrieved: 10.02.2019
    https://confluence.ecmwf.int/display/CKB/ERA5%3A+surface+elevation+and+orography

    """
    g0 = 9.80665
    z = ds["z"]
    if "time" in z.coords:
        z = z.isel(time=0, drop=True)
    ds["height"] = z / g0
    ds = ds.drop_vars("z")
    return ds


def _rename_and_clean_coords(ds, add_lon_lat=True):
    """
    Rename 'longitude' and 'latitude' columns to 'x' and 'y' and fix roundings.

    Optionally (add_lon_lat, default:True) preserves latitude and
    longitude columns as 'lat' and 'lon'.
    """
    ds = ds.rename({"longitude": "x", "latitude": "y"})
    if "valid_time" in ds.sizes:
        ds = ds.rename({"valid_time": "time"}).unify_chunks()
    # round coords since cds coords are float32 which would lead to mismatches
    ds = ds.assign_coords(
        x=np.round(ds.x.astype(float), 5), y=np.round(ds.y.astype(float), 5)
    )
    ds = maybe_swap_spatial_dims(ds)
    if add_lon_lat:
        ds = ds.assign_coords(lon=ds.coords["x"], lat=ds.coords["y"])
    ds = ds.drop_vars(["expver", "number"], errors="ignore")

    return ds


def get_data_wind(retrieval_params):
    """
    Get wind data for given retrieval parameters.
    """
    ds = retrieve_data(
        variable=[
            "10m_u_component_of_wind",
            "10m_v_component_of_wind",
            "100m_u_component_of_wind",
            "100m_v_component_of_wind",
            "forecast_surface_roughness",
        ],
        **retrieval_params,
    )
    ds = _rename_and_clean_coords(ds)

    for h in [10, 100]:
        ds[f"wnd{h}m"] = sqrt(ds[f"u{h}"] ** 2 + ds[f"v{h}"] ** 2).assign_attrs(
            units=ds[f"u{h}"].attrs["units"], long_name=f"{h} metre wind speed"
        )
    ds["wnd_shear_exp"] = (
        np.log(ds["wnd10m"] / ds["wnd100m"]) / np.log(10 / 100)
    ).assign_attrs(units="", long_name="wind shear exponent")

    # span the whole circle: 0 is north, π/2 is east, -π is south, 3π/2 is west
    azimuth = arctan2(ds["u100"], ds["v100"])
    ds["wnd_azimuth"] = azimuth.where(azimuth >= 0, azimuth + 2 * np.pi)

    ds = ds.drop_vars(["u100", "v100", "u10", "v10", "wnd10m"])
    ds = ds.rename({"fsr": "roughness"})

    return ds


def sanitize_wind(ds):
    """
    Sanitize retrieved wind data.
    """
    ds["roughness"] = ds["roughness"].where(ds["roughness"] >= 0.0, 2e-4)
    return ds


def get_data_influx(retrieval_params):
    """
    Get influx data for given retrieval parameters.
    """
    ds = retrieve_data(
        variable=[
            "surface_net_solar_radiation",
            "surface_solar_radiation_downwards",
            "toa_incident_solar_radiation",
            "total_sky_direct_solar_radiation_at_surface",
        ],
        **retrieval_params,
    )

    ds = _rename_and_clean_coords(ds)

    ds = ds.rename({"fdir": "influx_direct", "tisr": "influx_toa"})
    ds["albedo"] = (
        ((ds["ssrd"] - ds["ssr"]) / ds["ssrd"].where(ds["ssrd"] != 0))
        .fillna(0.0)
        .assign_attrs(units="(0 - 1)", long_name="Albedo")
    )
    ds["influx_diffuse"] = (ds["ssrd"] - ds["influx_direct"]).assign_attrs(
        units="J m**-2", long_name="Surface diffuse solar radiation downwards"
    )
    ds = ds.drop_vars(["ssrd", "ssr"])

    # Convert from energy to power J m**-2 -> W m**-2 and clip negative fluxes
    for a in ("influx_direct", "influx_diffuse", "influx_toa"):
        ds[a] = ds[a] / (60.0 * 60.0)
        ds[a].attrs["units"] = "W m**-2"

    # ERA5 variables are mean values for previous hour, i.e. 13:01 to 14:00 are labelled as "14:00"
    # account by calculating the SolarPosition for the center of the interval for aggregation happens
    # see https://github.com/PyPSA/atlite/issues/158
    # Do not show DeprecationWarning from new SolarPosition calculation (#199)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        time_shift = pd.to_timedelta("-30 minutes")
        sp = SolarPosition(ds, time_shift=time_shift)
    sp = sp.rename({v: f"solar_{v}" for v in sp.data_vars})

    ds = xr.merge([ds, sp])

    return ds


def sanitize_influx(ds):
    """
    Sanitize retrieved influx data.
    """
    for a in ("influx_direct", "influx_diffuse", "influx_toa"):
        ds[a] = ds[a].clip(min=0.0)
    return ds


def get_data_temperature(retrieval_params):
    """
    Get wind temperature for given retrieval parameters.
    """
    ds = retrieve_data(
        variable=[
            "2m_temperature",
            "soil_temperature_level_4",
            "2m_dewpoint_temperature",
        ],
        **retrieval_params,
    )

    ds = _rename_and_clean_coords(ds)
    ds = ds.rename(
        {
            "t2m": "temperature",
            "stl4": "soil temperature",
            "d2m": "dewpoint temperature",
        }
    )

    return ds


def get_data_runoff(retrieval_params):
    """
    Get runoff data for given retrieval parameters.
    """
    ds = retrieve_data(variable=["runoff"], **retrieval_params)

    ds = _rename_and_clean_coords(ds)
    ds = ds.rename({"ro": "runoff"})

    return ds


def sanitize_runoff(ds):
    """
    Sanitize retrieved runoff data.
    """
    ds["runoff"] = ds["runoff"].clip(min=0.0)
    return ds


def get_data_height(retrieval_params):
    """
    Get height data for given retrieval parameters.
    """
    ds = retrieve_data(variable="geopotential", **retrieval_params)

    ds = _rename_and_clean_coords(ds)
    ds = _add_height(ds)

    return ds


def _area(coords):
    # North, West, South, East. Default: global
    x0, x1 = coords["x"].min().item(), coords["x"].max().item()
    y0, y1 = coords["y"].min().item(), coords["y"].max().item()
    return [y1, x0, y0, x1]


def retrieval_times(coords, static=False, monthly_requests=False):
    """
    Get list of retrieval cdsapi arguments for time dimension in coordinates.

    If static is False, this function creates a query for each month and year
    in the time axis in coords. This ensures not running into size query limits
    of the cdsapi even with very (spatially) large cutouts.
    If static is True, the function return only one set of parameters
    for the very first time point.

    Parameters
    ----------
    coords : atlite.Cutout.coords
    static : bool, optional
    monthly_requests : bool, optional
        If True, the data is requested on a monthly basis. This is useful for
        large cutouts, where the data is requested in smaller chunks. The
        default is False

    Returns
    -------
    list of dicts witht retrieval arguments

    """
    time = coords["time"].to_index()
    if static:
        return {
            "year": str(time[0].year),
            "month": str(time[0].month),
            "day": str(time[0].day),
            "time": time[0].strftime("%H:00"),
        }

    # Prepare request for all months and years
    times = []
    for year in time.year.unique():
        t = time[time.year == year]
        if monthly_requests:
            for month in t.month.unique():
                query = {
                    "year": str(year),
                    "month": str(month),
                    "day": list(t[t.month == month].day.unique()),
                    "time": ["%02d:00" % h for h in t[t.month == month].hour.unique()],
                }
                times.append(query)
        else:
            query = {
                "year": str(year),
                "month": list(t.month.unique()),
                "day": list(t.day.unique()),
                "time": ["%02d:00" % h for h in t.hour.unique()],
            }
            times.append(query)
    return times


def noisy_unlink(path):
    """
    Delete file at given path.
    """
    logger.debug(f"Deleting file {path}")
    try:
        os.unlink(path)
    except PermissionError:
        logger.error(f"Unable to delete file {path}, as it is still in use.")


def retrieve_data(product, chunks=None, tmpdir=None, lock=None, **updates):
    """
    Download data like ERA5 from the Climate Data Store (CDS).

    If you want to track the state of your request go to
    https://cds-beta.climate.copernicus.eu/requests?tab=all
    """
    request = {"product_type": "reanalysis", "format": "netcdf"}
    request.update(updates)

    assert {"year", "month", "variable"}.issubset(
        request
    ), "Need to specify at least 'variable', 'year' and 'month'"

    client = cdsapi.Client(
        info_callback=logger.debug, debug=logging.DEBUG >= logging.root.level
    )
    result = client.retrieve(product, request)

    if lock is None:
        lock = nullcontext()

    with lock:
        fd, target = mkstemp(suffix=".nc", dir=tmpdir)
        os.close(fd)

        # Inform user about data being downloaded as "* variable (year-month)"
        timestr = f"{request['year']}-{request['month']}"
        variables = atleast_1d(request["variable"])
        varstr = "\n\t".join([f"{v} ({timestr})" for v in variables])
        logger.info(f"CDS: Downloading variables\n\t{varstr}\n")
        result.download(target)

    ds = xr.open_dataset(target, chunks=chunks or {})
    if tmpdir is None:
        logger.debug(f"Adding finalizer for {target}")
        weakref.finalize(ds._file_obj._manager, noisy_unlink, target)

    return ds


def get_data(
    cutout,
    feature,
    tmpdir,
    lock=None,
    monthly_requests=False,
    concurrent_requests=False,
    **creation_parameters,
):
    """
    Retrieve data from ECMWFs ERA5 dataset (via CDS).

    This front-end function downloads data for a specific feature and formats
    it to match the given Cutout.

    Parameters
    ----------
    cutout : atlite.Cutout
    feature : str
        Name of the feature data to retrieve. Must be in
        `atlite.datasets.era5.features`
    tmpdir : str/Path
        Directory where the temporary netcdf files are stored.
    monthly_requests : bool, optional
        If True, the data is requested on a monthly basis in ERA5. This is useful for
        large cutouts, where the data is requested in smaller chunks. The
        default is False
    concurrent_requests : bool, optional
        If True, the monthly data requests are posted concurrently.
        Only has an effect if `monthly_requests` is True.
    **creation_parameters :
        Additional keyword arguments. The only effective argument is 'sanitize'
        (default True) which sets sanitization of the data on or off.

    Returns
    -------
    xarray.Dataset
        Dataset of dask arrays of the retrieved variables.

    """
    coords = cutout.coords

    sanitize = creation_parameters.get("sanitize", True)

    retrieval_params = {
        "product": "reanalysis-era5-single-levels",
        "area": _area(coords),
        "chunks": cutout.chunks,
        "grid": [cutout.dx, cutout.dy],
        "tmpdir": tmpdir,
        "lock": lock,
    }

    func = globals().get(f"get_data_{feature}")
    sanitize_func = globals().get(f"sanitize_{feature}")

    logger.info(f"Requesting data for feature {feature}...")

    def retrieve_once(time):
        ds = func({**retrieval_params, **time})
        if sanitize and sanitize_func is not None:
            ds = sanitize_func(ds)
        return ds

    if feature in static_features:
        return retrieve_once(retrieval_times(coords, static=True)).squeeze()

    time_chunks = retrieval_times(coords, monthly_requests=monthly_requests)
    if concurrent_requests:
        delayed_datasets = [delayed(retrieve_once)(chunk) for chunk in time_chunks]
        datasets = compute(*delayed_datasets)
    else:
        datasets = map(retrieve_once, time_chunks)

    return xr.concat(datasets, dim="time").sel(time=coords["time"])



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Management of data retrieval and structure.
"""

import logging
import os
from functools import wraps
from shutil import rmtree
from tempfile import mkdtemp, mkstemp

import pandas as pd
import xarray as xr
from dask import compute, delayed
from dask.diagnostics import ProgressBar
from dask.utils import SerializableLock
from numpy import atleast_1d

from atlite.datasets import modules as datamodules

logger = logging.getLogger(__name__)


def get_features(
    cutout,
    module,
    features,
    tmpdir=None,
    monthly_requests=False,
    concurrent_requests=False,
):
    """
    Load the feature data for a given module.

    This get the data for a set of features from a module. All modules
    in `atlite.datasets` are allowed.
    """
    parameters = cutout.data.attrs
    lock = SerializableLock()
    datasets = []
    get_data = datamodules[module].get_data

    for feature in features:
        feature_data = delayed(get_data)(
            cutout,
            feature,
            tmpdir=tmpdir,
            lock=lock,
            monthly_requests=monthly_requests,
            concurrent_requests=concurrent_requests,
            **parameters,
        )
        datasets.append(feature_data)

    datasets = compute(*datasets)

    ds = xr.merge(datasets, compat="equals")
    for v in ds:
        ds[v].attrs["module"] = module
        fd = datamodules[module].features.items()
        ds[v].attrs["feature"] = [k for k, l in fd if v in l].pop()
    return ds


def available_features(module=None):
    """
    Inspect the available features of all or a selection of modules.

    Parameters
    ----------
    module : str/list, optional
        Module name(s) which to inspect. The default None will result in all
        modules

    Returns
    -------
    pd.Series
        A Series of all variables. The MultiIndex indicated which module
        provides the variable and with which feature name the variable can be
        obtained.

    """
    features = {name: m.features for name, m in datamodules.items()}
    features = (
        pd.DataFrame(features)
        .unstack()
        .dropna()
        .rename_axis(index=["module", "feature"])
        .rename("variables")
    )
    if module is not None:
        features = features.reindex(atleast_1d(module), level="module")
    return features.explode()


def non_bool_dict(d):
    """
    Convert bool to int for netCDF4 storing.
    """
    return {k: v if not isinstance(v, bool) else int(v) for k, v in d.items()}


def maybe_remove_tmpdir(func):
    """Use this wrapper to make tempfile deletion compatible with windows machines."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get("tmpdir", None):
            res = func(*args, **kwargs)
        else:
            kwargs["tmpdir"] = mkdtemp()
            try:
                res = func(*args, **kwargs)
            finally:
                rmtree(kwargs["tmpdir"])
        return res

    return wrapper


@maybe_remove_tmpdir
def cutout_prepare(
    cutout,
    features=None,
    tmpdir=None,
    overwrite=False,
    compression={"zlib": True, "complevel": 9, "shuffle": True},
    show_progress=False,
    dask_kwargs=None,
    monthly_requests=False,
    concurrent_requests=False,
):
    """
    Prepare all or a selection of features in a cutout.

    This function loads the feature data of a cutout, e.g. influx or runoff.
    When not specifying the `feature` argument, all available features will be
    loaded. The function compares the variables which are already included in
    the cutout with the available variables of the modules specified by the
    cutout. It detects missing variables and stores them into the netcdf file
    of the cutout.


    Parameters
    ----------
    cutout : atlite.Cutout
    features : str/list, optional
        Feature(s) to be prepared. The default slice(None) results in all
        available features.
    tmpdir : str/Path, optional
        Directory in which temporary files (for example retrieved ERA5 netcdf
        files) are stored. If set, the directory will not be deleted and the
        intermediate files can be examined.
    overwrite : bool, optional
        Whether to overwrite variables which are already included in the
        cutout. The default is False.
    compression : None/dict, optional
        Compression level to use for all features which are being prepared.
        The compression is handled via xarray.Dataset.to_netcdf(...), for details see:
        https://docs.xarray.dev/en/stable/generated/xarray.Dataset.to_netcdf.html .
        To efficiently reduce cutout sizes, specify the number of 'least_significant_digits': n here.
        To disable compression, set "complevel" to None.
        Default is {'zlib': True, 'complevel': 9, 'shuffle': True}.
    show_progress : bool, optional
        If True, a progress bar is shown. The default is False.
    dask_kwargs : dict, default {}
        Dict with keyword arguments passed to `dask.compute`.
    monthly_requests : bool, optional
        If True, the data is requested on a monthly basis in ERA5. This is useful for
        large cutouts, where the data is requested in smaller chunks. The
        default is False
    concurrent_requests : bool, optional
        If True, the monthly data requests are posted concurrently.
        Only has an effect if `monthly_requests` is True. The default is False.

    Returns
    -------
    cutout : atlite.Cutout
        Cutout with prepared data. The variables are stored in `cutout.data`.

    """
    if dask_kwargs is None:
        dask_kwargs = {}

    if cutout.prepared and not overwrite:
        logger.info("Cutout already prepared.")
        return cutout

    logger.info(f"Storing temporary files in {tmpdir}")

    modules = atleast_1d(cutout.module)
    features = atleast_1d(features) if features else slice(None)
    prepared = set(atleast_1d(cutout.data.attrs["prepared_features"]))

    # target is series of all available variables for given module and features
    target = available_features(modules).loc[:, features].drop_duplicates()

    for module in target.index.unique("module"):
        missing_vars = target[module]
        if not overwrite:
            missing_vars = missing_vars[lambda v: ~v.isin(cutout.data)]
        if missing_vars.empty:
            continue
        logger.info(f"Calculating and writing with module {module}:")
        missing_features = missing_vars.index.unique("feature")
        ds = get_features(
            cutout,
            module,
            missing_features,
            tmpdir=tmpdir,
            monthly_requests=monthly_requests,
            concurrent_requests=concurrent_requests,
        )
        prepared |= set(missing_features)

        cutout.data.attrs.update(dict(prepared_features=list(prepared)))
        attrs = non_bool_dict(cutout.data.attrs)
        attrs.update(ds.attrs)

        # Add optional compression to the newly prepared features
        if compression:
            for v in missing_vars:
                ds[v].encoding.update(compression)

        ds = cutout.data.merge(ds[missing_vars.values]).assign_attrs(**attrs)

        # write data to tmp file, copy it to original data, this is much safer
        # than appending variables
        directory, filename = os.path.split(str(cutout.path))
        fd, tmp = mkstemp(suffix=filename, dir=directory)
        os.close(fd)

        logger.debug("Writing cutout to file...")
        # Delayed writing for large cutout
        # cf. https://stackoverflow.com/questions/69810367/python-how-to-write-large-netcdf-with-xarray
        write_job = ds.to_netcdf(tmp, compute=False)
        if show_progress:
            with ProgressBar(minimum=2):
                write_job.compute(**dask_kwargs)
        else:
            write_job.compute(**dask_kwargs)
        if cutout.path.exists():
            cutout.data.close()
            cutout.path.unlink()
        os.rename(tmp, cutout.path)

        cutout.data = xr.open_dataset(cutout.path, chunks=cutout.chunks)

    return cutout



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Functions for use in conjunction with csp data generation.
"""

import logging

import numpy as np
from dask.array import radians, sin

from atlite.pv.solar_position import SolarPosition

logger = logging.getLogger(__name__)


def calculate_dni(ds, solar_position=None, altitude_threshold=3.75):
    """
    Calculate DNI on a perpendicular plane.

    Calculate the Direct Normal Irradiance (DNI) on a plane perpendicular to the solar
    irradiance based on solar altitude and direct solar influx on a horizontal plane.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing the direct influx (influx_direct) into a horizontal plane.
    solar_position : xarray.Dataset (optional)
        solar_position containing a solar 'altitude' (in rad, 0 to pi/2) for the 'ds' dataset.
        Is calculated using atlite.pv.SolarPosition if omitted.
    altitude_threshold : float (default: 3.75 degrees)
        Threshold for solar altitude in degrees. Values in range (0, altitude_threshold]
        will be set to the altitude_threshold to avoid numerical issues when dividing by
        the sine of very low solar altitude.
        The default values '3.75 deg' corresponds to
        the solar altitude traversed by the sun within about 15 minutes in a location with
        maximum solar altitude of 60 deg and 10h day time.

    """
    if solar_position is None:
        solar_position = SolarPosition(ds)

    # solar altitude expected in rad, convert degrees (easier to specifcy) to match
    altitude_threshold = radians(altitude_threshold)

    # Sanitation of altitude values:
    # Prevent high calculated DNI values during low solar altitudes (sunset / dawn)
    # where sin(<low altitude>) results in a very low denominator in the DNI calculation
    altitude = solar_position["altitude"]
    altitude = altitude.where(lambda x: x > 0, np.nan)
    altitude = altitude.where(lambda x: x > altitude_threshold, altitude_threshold)

    # Calculate DNI and remove NaNs introduced during altitude sanitation
    # DNI is determined either by dividing by cos(azimuth) or sin(altitude)
    dni = ds["influx_direct"] / sin(altitude)

    return dni



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Functions for use in conjunction with wind data generation.
"""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from typing import Literal


def extrapolate_wind_speed(
    ds: xr.Dataset,
    to_height: int | float,
    from_height: int | None = None,
    method: Literal["logarithmic", "power"] = "logarithmic",
) -> xr.DataArray:
    """
    Extrapolate the wind speed from a given height above ground to another.

    If ds already contains a key refering to wind speeds at the desired to_height,
    no conversion is done and the wind speeds are directly returned.

    Extrapolation of the wind speed can either use the "logarithmic" law as
    described in [1]_ or the "power" law as described in [2]_. See also discussion
    in GH issue: https://github.com/PyPSA/atlite/issues/231 .

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing the wind speed time-series at 'from_height' with key
        'wnd{height:d}m' and the surface orography with key 'roughness' at the
        geographic locations of the wind speeds.
    to_height : int|float
        Height (m) to which the wind speeds are extrapolated to.
    from_height : int, optional
        Height (m) from which the wind speeds are interpolated to 'to_height'.
        If not provided, the closest height to 'to_height' is selected.
    method : {"logarithmic", "power"}
        Method to use for extra/interpolating wind speeds

    Returns
    -------
    da : xarray.DataArray
        DataArray containing the extrapolated wind speeds. Name of the DataArray
        is 'wnd{to_height:d}'.

    Raises
    ------
    RuntimeError
        If the cutout is missing the data for the chosen `method`

    References
    ----------
    .. [1] Equation (2) in Andresen, G. et al (2015): 'Validation of Danish
       wind time series from a new global renewable energy atlas for energy
       system analysis'.

    .. [2] Gualtieri, G. (2021): 'Reliability of ERA5 Reanalysis Data for
       Wind Resource Assessment: A Comparison against Tall Towers'
       https://doi.org/10.3390/en14144169 .
    """
    # Fast lane
    to_name = f"wnd{int(to_height):0d}m"
    if to_name in ds:
        return ds[to_name]

    if from_height is None:
        # Determine closest height to to_name
        heights = np.asarray([int(s[3:-1]) for s in ds if re.match(r"wnd\d+m", s)])

        if len(heights) == 0:
            raise AssertionError("Wind speed is not in dataset")

        from_height = heights[np.argmin(np.abs(heights - to_height))]

    from_name = f"wnd{int(from_height):0d}m"

    if method == "logarithmic":
        try:
            roughness = ds["roughness"]
        except KeyError:
            raise RuntimeError(
                "The logarithmic interpolation method requires surface roughness (roughness);\n"
                "make sure you choose a compatible dataset like ERA5"
            )
        wnd_spd = ds[from_name] * (
            np.log(to_height / roughness) / np.log(from_height / roughness)
        )
        method_desc = "logarithmic method with roughness"
    elif method == "power":
        try:
            wnd_shear_exp = ds["wnd_shear_exp"]
        except KeyError:
            raise RuntimeError(
                "The power law interpolation method requires a wind shear exponent (wnd_shear_exp);\n"
                "make sure you choose a compatible dataset like ERA5 and update your cutout"
            )
        wnd_spd = ds[from_name] * (to_height / from_height) ** wnd_shear_exp
        method_desc = "power method with wind shear exponent"
    else:
        raise ValueError(
            f"Interpolation method must be 'logarithmic' or 'power', "
            f" but is: {method}"
        )

    wnd_spd.attrs.update(
        {
            "long name": (
                f"extrapolated {to_height} m wind speed using {method_desc} "
                f" and {from_height} m wind speed"
            ),
            "units": "m s**-1",
        }
    )

    return wnd_spd.rename(to_name)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Created on Wed May  6 15:23:13 2020.

@author: fabian
"""

# IDEAS for tests

import functools
import warnings

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import rasterio as rio
import rasterio.warp
import xarray as xr
from numpy import allclose, isclose
from shapely.geometry import box
from xarray.testing import assert_allclose, assert_equal

from atlite import Cutout
from atlite.gis import (
    ExclusionContainer,
    pad_extent,
    padded_transform_and_shape,
    regrid,
    shape_availability,
)

TIME = "2013-01-01"
X0 = -4.0
Y0 = 56.0
X1 = 1.5
Y1 = 61.0
raster_clip = 0.25  # this rastio is excluded (True) in the raster


@pytest.fixture(scope="session")
def ref(cutouts_path):
    tmp_path = cutouts_path / "creation_ref.nc"
    return Cutout(path=tmp_path, module="era5", bounds=(X0, Y0, X1, Y1), time=TIME)


@pytest.fixture(scope="session")
def geometry(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("geometries")
    geometry = gpd.GeoSeries(
        [box(X0 / 2 + X1 / 2, Y0 / 2 + Y1 / 2, X1, Y1)],
        crs="EPSG:4326",
        index=[0],
        name="boxes",
    )
    geometry = geometry.to_frame().set_geometry("boxes")

    path = tmp_path / "geometry.gpkg"

    geometry.to_file(path, driver="GPKG")

    return path


@pytest.fixture(scope="session")
def raster(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("rasters")
    bounds = (X0, Y0, X1, Y1)  # same as in test_gis.py
    res = 0.01
    transform, shape = padded_transform_and_shape(bounds, res)
    mask = np.random.rand(*shape) < raster_clip
    mask = mask.astype(rio.int32)
    path = tmp_path / "raster.tif"
    with rio.open(
        path,
        "w",
        driver="GTiff",
        transform=transform,
        crs=4326,
        width=shape[1],
        height=shape[0],
        count=1,
        dtype=mask.dtype,
    ) as dst:
        dst.write(mask, indexes=1)
    return path


@pytest.fixture(scope="session")
def raster_reproject(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("rasters")
    bounds = rio.warp.transform_bounds(4326, 3035, X0, Y0, X1, Y1)
    res = 1000
    transform, shape = padded_transform_and_shape(bounds, res)
    mask = np.random.rand(*shape) < raster_clip
    mask = mask.astype(rio.int32)
    path = tmp_path / "raster_reproject.tif"
    with rio.open(
        path,
        "w",
        driver="GTiff",
        transform=transform,
        crs=3035,
        width=shape[1],
        height=shape[0],
        count=1,
        dtype=mask.dtype,
    ) as dst:
        dst.write(mask, indexes=1)
    return path


@pytest.fixture(scope="session")
def raster_codes(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("rasters")
    bounds = (X0, Y0, X1, Y1)  # same as in test_gis.py
    res = 0.01
    transform, shape = padded_transform_and_shape(bounds, res)
    mask = (np.random.rand(*shape) * 100).astype(int)
    mask = mask.astype(rio.int32)
    path = tmp_path / "raster_codes.tif"
    with rio.open(
        path,
        "w",
        driver="GTiff",
        transform=transform,
        crs=4326,
        width=shape[1],
        height=shape[0],
        count=1,
        dtype=mask.dtype,
    ) as dst:
        dst.write(mask, indexes=1)
    return path


def test_exclusioncontainer_repr(ref):
    """
    Test ExclusionContainer.__repr__.
    """
    excluder = ExclusionContainer(ref.crs, res=0.01)
    assert "Exclusion Container" in repr(excluder)


def test_open_closed_checks(ref, geometry, raster):
    """
    Test atlite.ExclusionContainer(...) file open/closed checks for
    plausibility.

    C.f. GH issue #225.
    """
    res = 0.01
    excluder = ExclusionContainer(ref.crs, res=res)

    # Without raster/shapes, both should evaluate to True
    assert excluder.all_closed and excluder.all_open

    # First add geometries, than raster
    excluder.add_geometry(geometry)
    assert excluder.all_closed and not excluder.all_open

    # Check if still works with 2nd geometry
    excluder.add_geometry(geometry)
    assert excluder.all_closed and not excluder.all_open

    excluder.add_raster(raster)
    assert excluder.all_closed and not excluder.all_open

    # Check if still works with 2nd raster
    excluder.add_raster(raster)
    assert excluder.all_closed and not excluder.all_open

    excluder.open_files()
    assert not excluder.all_closed and excluder.all_open

    # First add raster, then geometries
    excluder = ExclusionContainer(ref.crs, res=res)

    excluder.add_raster(raster)
    assert excluder.all_closed and not excluder.all_open

    # 2nd raster
    excluder.add_raster(raster)
    assert excluder.all_closed and not excluder.all_open

    excluder.add_geometry(geometry)
    assert excluder.all_closed and not excluder.all_open

    # 2nd geometry
    excluder.add_geometry(geometry)
    assert excluder.all_closed and not excluder.all_open

    excluder.open_files()
    assert not excluder.all_closed and excluder.all_open


def test_area(ref):
    """
    Test the area of the cutout.
    """
    area = ref.area(crs=3035)
    assert isinstance(area, xr.DataArray)
    assert area.dims == ("y", "x")

    # now test with a different crs
    with pytest.warns(UserWarning):
        assert ref.area().sum().item() == (X1 - X0 + ref.dx) * (Y1 - Y0 + ref.dy)


def test_transform():
    """
    Test the affine transform.

    It always has to point to cell origin.
    """
    cutout = Cutout(
        path="resolution",
        module="era5",
        time=slice("2013-01-01", "2013-01-01"),
        x=slice(X0, X1),
        y=slice(Y0, Y1),
    )
    assert cutout.transform * (0.5, 0.5) == (X0, Y0)
    assert cutout.transform * cutout.shape[::-1] == (
        X1 + cutout.dx / 2,
        Y1 + cutout.dy / 2,
    )


def test_grid_coords(ref):
    gcoords = ref.grid[["x", "y"]]
    spatial = ref.data.stack(spatial=["y", "x"])["spatial"].data
    spatial = np.array([[s[1], s[0]] for s in spatial])
    np.testing.assert_equal(gcoords, spatial)


def test_extent(ref):
    pad = 0.25 / 2
    np.testing.assert_array_equal(ref.extent, [X0 - pad, X1 + pad, Y0 - pad, Y1 + pad])


# Note that bounds is the same as extent but in different order.
def test_bounds(ref):
    np.testing.assert_array_equal(ref.bounds, ref.grid.total_bounds)


def test_regrid():
    """
    Test the atlite.gis.regrid function with average resampling.
    """
    # define blocks
    A = 0.25
    B = 0.5
    C = 0.3
    D = 0.1
    ones = np.ones((4, 4))
    fine = np.block([[ones * A, ones * B], [ones * C, ones * D]])
    # add coordinates
    finecoords = np.arange(0.5, 8, 1)
    fine = xr.DataArray(fine, coords=[("y", finecoords), ("x", finecoords)])

    coarsecoords = np.arange(2, 8, 4)
    coarse = xr.DataArray(np.nan, coords=[("y", coarsecoords), ("x", coarsecoords)])

    # apply average resampling
    res = regrid(fine, coarse.x, coarse.y, resampling=5)
    target = np.array([[A, B], [C, D]])
    assert allclose(res, target)
    assert (coarse.x == res.x).all() and (coarse.y == res.y).all()

    # now test multiple layers

    fine = xr.concat([fine] * 10, pd.Index(range(10), name="z"))
    res = regrid(fine, coarse.x, coarse.y, resampling=5)
    target = np.stack([np.array([[A, B], [C, D]])] * 10)
    assert allclose(res, target)
    assert (coarse.x == res.x).all() and (coarse.y == res.y).all()

    # now let the target grid cover a subarea of the original
    fine = fine.sel(z=0, drop=True)
    coarsecoords = np.arange(1, 6, 2)
    coarse = xr.DataArray(np.nan, coords=[("y", coarsecoords), ("x", coarsecoords)])

    # apply average resampling
    res = regrid(fine, coarse.x, coarse.y, resampling=5)
    target = np.array([[A, A, B], [A, A, B], [C, C, D]])
    assert allclose(res, target)
    assert (coarse.x == res.x).all() and (coarse.y == res.y).all()


def test_pad_extent():
    """
    Test whether padding works with arrays of dimension > 2.
    """
    src = np.ones((3, 2))
    src_trans = rio.Affine(1, 0, 0, 0, 1, 0)
    dst_trans = rio.Affine(2, 0, 0, 0, 2, 0)
    crs = 4326

    padded, trans = pad_extent(src, src_trans, dst_trans, crs, crs)
    src = np.ones((1, 3, 2))
    padded_ndim, trans_ndim = pad_extent(src, src_trans, dst_trans, crs, crs)
    assert (padded_ndim[0] == padded).all()
    assert trans == trans_ndim

    # second check with large shape
    src = np.ones((1, 2, 3, 2))
    padded_ndim, trans_ndim = pad_extent(src, src_trans, dst_trans, crs, crs)
    assert (padded_ndim[0, 0] == padded).all()
    assert trans == trans_ndim

    # other way round, here it should not pad, since target resolution is lower
    padded_r, trans_r = pad_extent(src, dst_trans, src_trans, crs, crs)
    assert (padded_r == src).all()
    assert trans_r == dst_trans


def test_indicator_matrix(ref):
    # This should be the grid cell at the lower left corner
    cell = ref.grid.geometry[0]
    indicator = ref.indicatormatrix([cell])
    assert indicator[0, 0] == 1.0
    assert indicator.sum() == 1
    # This should be the grid cell at the lower left corner
    cell = ref.grid.geometry.iloc[-2]
    indicator = ref.indicatormatrix([cell])
    assert indicator[0, -2] == 1.0
    assert indicator.sum() == 1


def test_availability_matrix_flat(ref):
    """
    Indicator matrix and availability matrix for an empty excluder must be the
    same.
    """
    shapes = gpd.GeoSeries(
        [box(X0 + 1, Y0 + 1, X1 - 1, Y1 - 1)], crs=ref.crs
    ).rename_axis("shape")
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords["y"], ref.coords["x"]])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    ds = ref.availabilitymatrix(shapes, excluder)
    assert np.allclose(I, ds.sum("shape"))


def test_availability_matrix_flat_parallel(ref):
    """
    Same as `test_availability_matrix_flat` but parallel and without
    progressbar.
    """
    shapes = gpd.GeoSeries(
        [box(X0 + 1, Y0 + 1, X1 - 1, Y1 - 1)], crs=ref.crs
    ).rename_axis("shape")
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords["y"], ref.coords["x"]])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    ds = ref.availabilitymatrix(shapes, excluder, nprocesses=2)
    assert np.allclose(I, ds.sum("shape"))


def test_availability_matrix_flat_parallel_anonymous_function(ref, raster_codes):
    """
    Test availability matrix in parallel mode with a non-anonymous filter
    function.
    """
    shapes = gpd.GeoSeries(
        [box(X0 + 1, Y0 + 1, X1 - 1, Y1 - 1)], crs=ref.crs
    ).rename_axis("shape")
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords["y"], ref.coords["x"]])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    func = functools.partial(np.greater_equal, 20)
    excluder.add_raster(raster_codes, codes=func)
    ref.availabilitymatrix(shapes, excluder, nprocesses=2)


def test_availability_matrix_flat_wo_progressbar(ref):
    """
    Same as `test_availability_matrix_flat` but without progressbar.
    """
    shapes = gpd.GeoSeries(
        [box(X0 + 1, Y0 + 1, X1 - 1, Y1 - 1)], crs=ref.crs
    ).rename_axis("shape")
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords["y"], ref.coords["x"]])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    ds = ref.availabilitymatrix(shapes, excluder, disable_progressbar=True)
    assert np.allclose(I, ds.sum("shape"))


def test_availability_matrix_flat_parallel_wo_progressbar(ref):
    """
    Same as `test_availability_matrix_flat` but parallel and without
    progressbar.
    """
    shapes = gpd.GeoSeries(
        [box(X0 + 1, Y0 + 1, X1 - 1, Y1 - 1)], crs=ref.crs
    ).rename_axis("shape")
    I = ref.indicatormatrix(shapes).sum(0).reshape(ref.shape)
    I = xr.DataArray(I, coords=[ref.coords["y"], ref.coords["x"]])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    ds = ref.availabilitymatrix(
        shapes, excluder, nprocesses=2, disable_progressbar=True
    )
    assert np.allclose(I, ds.sum("shape"))


def test_shape_availability_area(ref):
    """
    Area of the mask and the shape must be close.
    """
    shapes = gpd.GeoSeries([box(X0 + 1, Y0 + 1, X1 - 1, Y1 - 1)], crs=ref.crs)
    res = 100
    excluder = ExclusionContainer(res=res)

    with pytest.raises(AssertionError):
        masked, transform = shape_availability(shapes, excluder)

    shapes = shapes.to_crs(3035)
    masked, transform = shape_availability(shapes, excluder)
    assert np.isclose(shapes.area, masked.sum() * res**2)

    masked2, transform2 = excluder.compute_shape_availability(shapes)
    assert (masked == masked2).all()
    assert transform == transform2

    masked3, transform3 = excluder.compute_shape_availability(shapes.to_frame())
    assert (masked == masked2).all()
    assert transform == transform2


def test_exclusioncontainer_geometries():
    crs = 3035
    exclude = gpd.GeoSeries([box(X0 / 2 + X1 / 2, Y0 / 2 + Y1 / 2, X1, Y1)], crs=crs)
    res = 0.01

    excluder = ExclusionContainer(crs, res=res)
    excluder.add_geometry(exclude, buffer=1)
    excluder.open_files()
    assert (excluder.geometries[0]["geometry"] != exclude).all()
    excluder.open_files()
    buffered = excluder.geometries[0]["geometry"]
    # open again and check that the buffer remains the same
    assert (excluder.geometries[0]["geometry"] == buffered).all()

    # should take GeoDataFrames and the result is the same
    excluder = ExclusionContainer(crs, res=res)
    excluder.add_geometry(exclude.to_frame("geometry"), buffer=1)
    excluder.open_files()
    assert (excluder.geometries[0]["geometry"] == buffered).all()


def test_shape_availability_exclude_geometry(ref):
    """
    When excluding the quarter of the geometry, the eligible area must be a
    forth.

    Test the inverted case too.
    """
    shapes = gpd.GeoSeries([box(X0, Y0, X1, Y1)], crs=ref.crs)
    exclude = gpd.GeoSeries(
        [box(X0 / 2 + X1 / 2, Y0 / 2 + Y1 / 2, X1, Y1)], crs=ref.crs
    )
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_geometry(exclude)
    masked, transform = shape_availability(shapes, excluder)
    area = shapes.geometry[0].area  # get area without warning
    assert isclose(3 * area / 4, masked.sum() * res**2)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_geometry(exclude, invert=True)
    masked, transform = shape_availability(shapes, excluder)
    area = shapes.geometry[0].area  # get area without warning
    assert isclose(area / 4, masked.sum() * res**2)


def test_shape_availability_exclude_raster(ref, raster):
    """
    When excluding the half of the geometry, the eligible area must be half.
    """
    shapes = gpd.GeoSeries([box(X0, Y0, X1, Y1)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster)
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    assert round(ratio, 2) == (1 - raster_clip)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, invert=True)
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    assert round(ratio, 2) == raster_clip

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, buffer=res)
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    # should be close to zero
    assert round(ratio, 2) < (1 - raster_clip)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, buffer=res, invert=True)
    masked, transform = shape_availability(shapes, excluder)
    ratio2 = masked.sum() / masked.size
    # for the case that we have more excluded area and this is buffered
    if raster_clip < 0.5:
        assert ratio >= ratio2
    else:
        assert ratio <= ratio2


def test_shape_availability_excluder_partial_overlap(ref, raster):
    """
    Test behavior, when a raster only overlaps half of the geometry.
    """
    bounds = X0 - 2, Y0, X0 + 2, Y1
    area = abs((bounds[2] - bounds[0]) * (bounds[3] - bounds[1]))
    shapes = gpd.GeoSeries([box(*bounds)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, codes=[0, 1])
    masked, transform = shape_availability(shapes, excluder)
    assert masked.sum() * (res**2) == area / 2

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, nodata=0)
    masked, transform = shape_availability(shapes, excluder)
    assert masked.sum() * (res**2) > area / 2

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, nodata=1)
    masked, transform = shape_availability(shapes, excluder)
    assert masked.sum() * (res**2) < area / 2


def test_shape_availability_excluder_raster_no_overlap(ref, raster):
    """
    Check if the allow_no_overlap flag works.
    """
    bounds = X0 - 10.0, Y0 - 10.0, X0 - 2.0, Y0 - 2.0
    area = abs((bounds[2] - bounds[0]) * (bounds[3] - bounds[1]))
    shapes = gpd.GeoSeries([box(*bounds)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster)
    with pytest.raises(ValueError):
        masked, transform = shape_availability(shapes, excluder)

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, allow_no_overlap=True)
    masked, transform = shape_availability(shapes, excluder)
    assert (masked == 0).all()

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, allow_no_overlap=True, codes=[1, 255], invert=True)
    masked, transform = shape_availability(shapes, excluder)
    assert masked.sum() * (res**2) == area

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster, allow_no_overlap=True, nodata=0)
    masked, transform = shape_availability(shapes, excluder)
    assert masked.sum() * (res**2) == area


def test_availability_matrix_rastered(ref, raster):
    """
    Availability matrix with a non-zero raster must have less available area
    than the Indicator matrix.
    """
    shapes = gpd.GeoSeries(
        [
            box(X0 + 1, Y0 + 1, X1 - 1, Y0 / 2 + Y1 / 2),
            box(X0 + 1, Y0 / 2 + Y1 / 2, X1 - 1, Y1 - 1),
        ],
        crs=ref.crs,
    ).rename_axis("shape")
    I = np.asarray(ref.indicatormatrix(shapes).todense())
    I = I.reshape(shapes.shape + ref.shape)
    I = xr.DataArray(I, coords=[shapes.index, ref.coords["y"], ref.coords["x"]])
    excluder = ExclusionContainer(ref.crs, res=0.01)
    excluder.add_raster(raster)
    ds = ref.availabilitymatrix(shapes, excluder)
    eligible_share = 1 - raster_clip

    assert isclose(I.sum() * eligible_share, ds.sum(), atol=5)
    assert_allclose(I.sum(["x", "y"]) * eligible_share, ds.sum(["x", "y"]), atol=5)

    excluder = ExclusionContainer(ref.crs, res=0.01)
    excluder.add_raster(raster)
    assert_equal(ds, ref.availabilitymatrix(shapes, excluder, 2))


def test_availability_matrix_rastered_repro(ref, raster_reproject):
    """
    Availability matrix with a non-zero raster must have less available area
    than the Indicator matrix.

    Test this with a raster of a different crs.
    """
    shapes = gpd.GeoSeries(
        [
            box(X0 + 1, Y0 + 1, X1 - 1, Y0 / 2 + Y1 / 2),
            box(X0 + 1, Y0 / 2 + Y1 / 2, X1 - 1, Y1 - 1),
        ],
        crs=ref.crs,
    ).rename_axis("shape")
    I = np.asarray(ref.indicatormatrix(shapes).todense())
    I = I.reshape(shapes.shape + ref.shape)
    I = xr.DataArray(I, coords=[shapes.index, ref.coords["y"], ref.coords["x"]])
    excluder = ExclusionContainer()
    excluder.add_raster(raster_reproject)
    ds = ref.availabilitymatrix(shapes, excluder)
    eligible_share = 1 - raster_clip

    assert isclose(I.sum() * eligible_share, ds.sum(), atol=5)
    assert_allclose(I.sum(["x", "y"]) * eligible_share, ds.sum(["x", "y"]), atol=5)


def test_shape_availability_exclude_raster_codes(ref, raster_codes):
    """
    Test exclusion of multiple raster codes.
    """
    shapes = gpd.GeoSeries([box(X0, Y0, X1, Y1)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster_codes, codes=range(20))
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    assert round(ratio, 1) == 0.8

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster_codes, codes=range(20), invert=True)
    masked, transform = shape_availability(shapes, excluder)
    ratio = masked.sum() / masked.size
    assert round(ratio, 1) == 0.2

    # test with a function
    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster_codes, codes=lambda x: x < 20, invert=True)
    masked, transform = shape_availability(shapes, excluder)
    assert ratio == masked.sum() / masked.size


def test_plot_shape_availability(ref, raster):
    """
    Test plotting of shape availability.
    """
    shapes = gpd.GeoSeries([box(X0, Y0, X1, Y1)], crs=ref.crs)
    res = 0.01

    excluder = ExclusionContainer(ref.crs, res=res)
    excluder.add_raster(raster)
    # disable UserWarning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        excluder.plot_shape_availability(shapes)



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Spyder Editor.

This is a temporary script file.
"""

import matplotlib.pyplot as plt

fig, ax = plt.subplots(figsize=(12, 5))
ax.axis("off")

linebreak = "\n    "

climatedata = [
    "Temperature",
    "Downward short-wave radiation",
    "Upward short-wave radiation",
    "Wind",
    "Runoff",
    "Surface roughness",
    "Height maps",
    "Soil temperature",
]
processeddata = [
    f"Wind power generation {linebreak}for a given turbine type",
    f"Solar PV power generation {linebreak}for a given panel type",
    "Solar thermal collector heat output",
    "Hydroelectric inflow (simplified)",
    f"Heating demand {linebreak}" "(based on the degree-day approximation)",
]

climatestr = "\n" + "\n\n".join([" ◦ " + s for s in climatedata]) + "\n"

processedstr = "\n" + "\n\n\n".join([" ◦ " + s for s in processeddata]) + "\n"

# defaults for boxes and arrows
kwargs = dict(verticalalignment="center", fontsize=14, color="#545454")
arrowkwargs = dict(
    head_width=0.2,
    width=0.13,
    head_length=0.05,
    edgecolor="white",
    length_includes_head=True,
    color="lightgray",
    alpha=1,
)
y = 0.5

# First arrow
ax.text(
    0.01, y, " Retrieve Data", fontsize=14, color="gray", verticalalignment="center"
)
ax.arrow(0.01, y, 0.14, 0.0, **arrowkwargs)

# First box
ax.text(
    0.17,
    y,
    climatestr,
    **kwargs,
    bbox=dict(facecolor="indianred", alpha=0.5, edgecolor="None", boxstyle="round"),
)

# Second arrow
ax.text(0.5, y, " Process Data", fontsize=14, color="gray", verticalalignment="center")
ax.arrow(0.5, y, 0.14, 0, **arrowkwargs)

# Second Box
ax.text(
    0.66,
    y,
    processedstr,
    **kwargs,
    bbox=dict(facecolor="olivedrab", alpha=0.5, edgecolor="None", boxstyle="round"),
)


fig.tight_layout(pad=0)
fig.savefig("workflow.png", dpi=150)



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Module for providing access to external ressources, like windturbine or pv
panel configurations.
"""

from __future__ import annotations

import json
import logging
import re
import warnings
from operator import itemgetter
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import requests
import yaml
from dask.array import radians
from scipy.signal import fftconvolve

from atlite.utils import arrowdict

logger = logging.getLogger(name=__name__)


RESOURCE_DIRECTORY = Path(__file__).parent / "resources"
WINDTURBINE_DIRECTORY = RESOURCE_DIRECTORY / "windturbine"
SOLARPANEL_DIRECTORY = RESOURCE_DIRECTORY / "solarpanel"
CSPINSTALLATION_DIRECTORY = RESOURCE_DIRECTORY / "cspinstallation"

if TYPE_CHECKING:
    from typing import TypedDict

    from typing_extensions import NotRequired

    class TurbineConfig(TypedDict):
        V: np.ndarray
        POW: np.ndarray
        P: float
        hub_height: float | int
        name: NotRequired[str]
        manufacturer: NotRequired[str]
        source: NotRequired[str]


def get_windturbineconfig(
    turbine: str | Path | dict, add_cutout_windspeed: bool = False
) -> TurbineConfig:
    """
    Load the wind 'turbine' configuration.

    Parameters
    ----------
    turbine : str or pathlib.Path or dict
        if str:
            The name of a preshipped turbine from alite.resources.windturbine .
            Alternatively, if a str starting with 'oedb:<name>' is passed the Open
            Energy Database is searched for a turbine with the matching '<name>'
            and if found that turbine configuration is used. See
            `atlite.resource.get_oedb_windturbineconfig(...)`
        if `pathlib.Path` is provided the configuration is read from this local
            path instead
        if dict:
            a user provided config dict. Needs to have the keys "POW", "V", "P", and
            "hub_height". Values for "POW" and "V" need to be list or np.ndarray with
            equal length.
    add_cutout_windspeed : bool
        If True and in case the power curve does not end with a zero, will add zero power
        output at the highest wind speed in the power curve. If False, a warning will be
        raised if the power curve does not have a cut-out wind speed.

    Returns
    -------
    config : dict
        Config with details on the turbine

    """
    assert isinstance(turbine, (str, Path, dict))

    if add_cutout_windspeed is False:
        msg = (
            "'add_cutout_windspeed' for wind turbine\npower curves will default to "
            "True in atlite relase v0.2.15."
        )
        warnings.warn(msg, FutureWarning)

    if isinstance(turbine, str) and turbine.startswith("oedb:"):
        conf = get_oedb_windturbineconfig(turbine[len("oedb:") :])

    elif isinstance(turbine, (str, Path)):
        if isinstance(turbine, str):
            turbine_path = windturbines[turbine.replace(".yaml", "")]

        elif isinstance(turbine, Path):
            turbine_path = turbine

        with open(turbine_path) as f:
            conf = yaml.safe_load(f)
            conf = dict(
                V=np.array(conf["V"]),
                POW=np.array(conf["POW"]),
                hub_height=conf["HUB_HEIGHT"],
                P=np.max(conf["POW"]),
            )

    elif isinstance(turbine, dict):
        conf = turbine

    return _validate_turbine_config_dict(conf, add_cutout_windspeed)


def get_solarpanelconfig(panel):
    """
    Load the 'panel'.yaml file from local disk and provide a solar panel dict.

    Parameters
    ----------
    panel : str or pathlib.Path
        if str is provided the name of a preshipped panel
            from alite.resources.solarpanel is expected.
        if `pathlib.Path` is provided the configuration
            is read from this local path instead

    Returns
    -------
    config : dict
        Config with details on the solarpanel

    """
    assert isinstance(panel, (str, Path))

    if isinstance(panel, str):
        panel_path = solarpanels[panel.replace(".yaml", "")]

    elif isinstance(panel, Path):
        panel_path = panel

    with open(panel_path) as f:
        conf = yaml.safe_load(f)

    return conf


def get_cspinstallationconfig(installation):
    """
    Load the 'installation'.yaml file from local disk to provide the system
    efficiencies.

    Parameters
    ----------
    installation : str or pathlib.Path
        if str is provided the name of a preshipped CSP installation
            from alite.resources.cspinstallation is expected.
        if `pathlib.Path` is provided the configuration
            is read from this local path instead

    Returns
    -------
    config : dict
        Config with details on the CSP installation.

    """
    assert isinstance(installation, (str, Path))

    if isinstance(installation, str):
        installation_path = cspinstallations[installation.replace(".yaml", "")]

    elif isinstance(installation, Path):
        installation_path = installation

    # Load and set expected index columns
    with open(installation_path) as f:
        config = yaml.safe_load(f)
    config["path"] = installation_path

    ## Convert efficiency dict to xr.DataArray and convert units to deg -> rad, % -> p.u.
    da = pd.DataFrame(config["efficiency"]).set_index(["altitude", "azimuth"])

    # Handle as xarray DataArray early - da will be 'return'-ed
    da = da.to_xarray()["value"]

    # Solar altitude + azimuth expected in deg for better readibility
    # calculations use solar position in rad
    # Convert da to new coordinates and drop old
    da = da.rename({"azimuth": "azimuth [deg]", "altitude": "altitude [deg]"})
    da = da.assign_coords(
        {
            "altitude": radians(da["altitude [deg]"]),
            "azimuth": radians(da["azimuth [deg]"]),
        }
    )
    da = da.swap_dims({"altitude [deg]": "altitude", "azimuth [deg]": "azimuth"})

    da = da.chunk("auto")

    # Efficiency unit from % to p.u.
    da /= 1.0e2

    config["efficiency"] = da

    return config


def solarpanel_rated_capacity_per_unit(panel):
    # unit is m^2 here

    if isinstance(panel, (str, Path)):
        panel = get_solarpanelconfig(panel)

    model = panel.get("model", "huld")
    if model == "huld":
        return panel["efficiency"]
    elif model == "bofinger":
        # one unit in the capacity layout is interpreted as one panel of a
        # capacity (A + 1000 * B + log(1000) * C) * 1000W/m^2 * (k / 1000)
        A, B, C = itemgetter("A", "B", "C")(panel)
        return (A + B * 1000.0 + C * np.log(1000.0)) * 1e3


def windturbine_rated_capacity_per_unit(turbine):
    if isinstance(turbine, (str, Path)):
        turbine = get_windturbineconfig(turbine)

    return turbine["P"]


def windturbine_smooth(turbine, params=None):
    """
    Smooth the powercurve in `turbine` with a gaussian kernel.

    Parameters
    ----------
    turbine : dict
        Turbine config with at least V and POW
    params : dict
        Allows adjusting fleet availability eta, mean Delta_v and
        stdev sigma. Defaults to values from Andresen's paper: 0.95,
        1.27 and 2.29, respectively.

    Returns
    -------
    turbine : dict
        Turbine config with a smoothed power curve

    References
    ----------
    G. B. Andresen, A. A. Søndergaard, M. Greiner, Validation of
    Danish wind time series from a new global renewable energy atlas
    for energy system analysis, Energy 93, Part 1 (2015) 1074–1088.

    """
    if params is None or params is True:
        params = {}

    eta = params.get("eta", 0.95)
    Delta_v = params.get("Delta_v", 1.27)
    sigma = params.get("sigma", 2.29)

    def kernel(v_0):
        # all velocities in m/s
        return (
            1.0
            / np.sqrt(2 * np.pi * sigma * sigma)
            * np.exp(-(v_0 - Delta_v) * (v_0 - Delta_v) / (2 * sigma * sigma))
        )

    def smooth(velocities, power):
        # interpolate kernel and power curve to the same, regular velocity grid
        velocities_reg = np.linspace(-50.0, 50.0, 1001)
        power_reg = np.interp(velocities_reg, velocities, power)
        kernel_reg = kernel(velocities_reg)

        # convolve power and kernel
        # the downscaling is necessary because scipy expects the velocity
        # increments to be 1., but here, they are 0.1
        convolution = 0.1 * fftconvolve(power_reg, kernel_reg, mode="same")

        # sample down so power curve doesn't get too long
        velocities_new = np.linspace(0.0, 35.0, 72)
        power_new = eta * np.interp(velocities_new, velocities_reg, convolution)

        return velocities_new, power_new

    turbine = turbine.copy()
    turbine["V"], turbine["POW"] = smooth(turbine["V"], turbine["POW"])
    turbine["P"] = np.max(turbine["POW"])

    if any(turbine["POW"][np.where(turbine["V"] == 0.0)] > 1e-2):
        logger.warning(
            "Oversmoothing detected with parameters eta=%f, Delta_v=%f, sigma=%f. "
            "Turbine generates energy at 0 m/s wind speeds.",
            eta,
            Delta_v,
            sigma,
        )

    return turbine


def _max_v_is_zero_pow(turbine):
    return np.any(turbine["POW"][turbine["V"] == turbine["V"].max()] == 0)


def _validate_turbine_config_dict(
    turbine: dict, add_cutout_windspeed: bool
) -> TurbineConfig:
    """
    Checks the turbine config dict format and power curve.

    Parameters
    ----------
    turbine : dict
        turbine configuration dict. Needs the keys "POW", "V", "P", and "hub_height".
        Values for "V" and "POW" need to be list or np.ndarray.
    add_cutout_windspeed : bool
        If True and in case the power curve does not end with a zero, will add zero power
        output at the highest wind speed in the power curve. If False, a warning will be
        raised if the power curve does not have a cut-out wind speed.

    Returns
    -------
    dict
        validated and potentially modified turbine config dict

    """
    if not all(key in turbine for key in ("POW", "V", "P", "hub_height")):
        err_msg = (
            "turbine config dict needs at least the following keys: ['POW', 'V', 'P', "
            f"'hub_height']\nbut are currently: {list(turbine.keys())}"
        )
        raise ValueError(err_msg)

    if not all(isinstance(turbine[p], (np.ndarray, list)) for p in ("POW", "V")):
        err_msg = "turbine entries 'POW' and 'V' must be np.ndarray or list"
        raise ValueError(err_msg)

    # convert lists from user provided turbine dicts to numpy arrays
    if any(isinstance(turbine[p], list) for p in ("POW", "V")):
        turbine["V"] = np.array(turbine["V"])
        turbine["POW"] = np.array(turbine["POW"])

    if len(turbine["POW"]) != len(turbine["V"]):
        err_msg = "turbine wind speed and power arrays do not have equal length."
        raise ValueError(err_msg)

    if not np.all(np.diff(turbine["V"]) >= 0):
        # This check is not strict as it uses `>=` instead of `>` and thus allows equal
        # wind speeds in the array. However, many power curves have two entries for the
        # same wind speed at the cut-in and cut-out speeds which would make them fail if
        # using `>` only.
        err_msg = (
            "wind speed 'V' in the turbine config dict is expected to be increasing, "
            f"but is currently not in ascending order:\n{turbine['V']}"
        )
        raise ValueError(err_msg)

    if add_cutout_windspeed is True and not _max_v_is_zero_pow(turbine):
        turbine["V"] = np.pad(turbine["V"], (0, 1), "maximum")
        turbine["POW"] = np.pad(turbine["POW"], (0, 1), "constant", constant_values=0)
        logger.info(
            "adding a cut-out wind speed to the turbine power curve at "
            f"V={turbine['V'][-1]} m/s."
        )

    if not _max_v_is_zero_pow(turbine):
        logger.warning(
            "The power curve does not have a cut-out wind speed, i.e. the power"
            " output corresponding to the\nhighest wind speed is not zero. You can"
            " either change the power curve manually or set\n"
            "'add_cutout_windspeed=True' in the Cutout.wind conversion method."
        )
    return turbine


def get_oedb_windturbineconfig(
    search: int | str | None = None, **search_params
) -> TurbineConfig:
    """
    Download a windturbine configuration from the OEDB database.

    Download the configuration of a windturbine model from the OEDB database
    into the local 'windturbine_dir'.
    The OEDB database can be viewed here:
    https://openenergy-platform.org/dataedit/view/supply/wind_turbine_library
    (2019-07-22)
    Only one turbine configuration is downloaded at a time, if the
    search parameters yield an ambigious result, no data is downloaded.

    Parameters
    ----------
    search : int|str
        Smart search parameter, if int use as model id, if str look in name or turbine_type
    **search_params : dict
        Recognized arguments are 'id', 'name', 'turbine_type' and 'manufacturer'

    Returns
    -------
    turbineconfig : dict
        The turbine configuration in the format from 'atlite.ressource.get_turbineconf(name)'.

    Example
    -------
    >>> get_oedb_windturbineconfig(10)
    {'V': ..., 'POW': ..., ...}

    >>> get_oedb_windturbineconfig(name="E-53/800", manufacturer="Enercon")
    {'V': ..., 'POW': ..., ...}

    """
    # Parse information of different allowed 'turbine' values
    if isinstance(search, int):
        search_params.setdefault("id", search)
        search = None

    # Retrieve and cache OEDB turbine data
    OEDB_URL = "https://openenergy-platform.org/api/v0/schema/supply/tables/wind_turbine_library/rows"

    # Cache turbine request locally
    global _oedb_turbines

    if _oedb_turbines is None:
        # Get the turbine list
        result = requests.get(OEDB_URL)

        # Convert JSON to dataframe for easier filtering
        # Only consider turbines with power curves available
        df = pd.DataFrame.from_dict(result.json())
        _oedb_turbines = df[df.has_power_curve]

    logger.info(
        "Searching turbine power curve in OEDB database using "
        + ", ".join(f"{k}='{v}'" for (k, v) in search_params.items())
        + "."
    )

    # Working copy
    df = _oedb_turbines
    selector = True
    if search is not None:
        selector &= df.name.str.contains(
            search, case=False
        ) | df.turbine_type.str.contains(search, case=False)
    if "id" in search_params:
        selector &= df.id == int(search_params["id"])
    if "name" in search_params:
        selector &= df.name.str.contains(search_params["name"], case=False)
    if "turbine_type" in search_params:
        selector &= df.turbine_type.str.contains(search_params["name"], case=False)
    if "manufacturer" in search_params:
        selector &= df.manufacturer.str.contains(
            search_params["manufacturer"], case=False
        )

    df = df.loc[selector]

    if len(df) < 1:
        raise RuntimeError("No turbine found.")
    elif len(df) > 1:
        raise RuntimeError(
            f"Provided information corresponds to {len(df)} turbines,"
            " use `id` for an unambiguous search.\n"
            + str(df[["id", "manufacturer", "turbine_type"]])
        )

    # Convert to series for simpliticty
    ds = df.iloc[0]

    # convert power from kW to MW
    power = np.array(json.loads(ds.power_curve_values)) / 1e3

    hub_height = ds.hub_height

    if not hub_height:
        hub_height = 100
        logger.warning(
            "No hub_height defined in dataset. Manual clean-up required."
            "Assuming a hub_height of 100m for now."
        )
    elif isinstance(hub_height, str):
        hub_heights = [float(t) for t in re.split(r"\s*;\s*", hub_height.strip()) if t]

        if len(hub_heights) > 1:
            hub_height = np.mean(hub_heights, dtype=int)
            logger.warning(
                "Multiple values for hub_height in dataset (%s). "
                "Manual clean-up required. Using the averge %dm for now.",
                hub_heights,
                hub_height,
            )
        else:
            hub_height = hub_heights[0]

    turbineconf = {
        "name": ds.turbine_type.strip(),
        "manufacturer": ds.manufacturer.strip(),
        "source": f"Original: {ds.source}. Via OEDB {OEDB_URL}",
        "hub_height": hub_height,
        "V": np.array(json.loads(ds.power_curve_wind_speeds)),
        "POW": power,
        "P": power.max(),
    }

    # Cache in windturbines
    global windturbines
    charmap = str.maketrans("/- ", "___")
    name = "{manufacturer}_{name}".format(**turbineconf).translate(charmap)
    windturbines[name] = turbineconf

    return turbineconf


# Global caches
_oedb_turbines = None
windturbines = arrowdict({p.stem: p for p in WINDTURBINE_DIRECTORY.glob("*.yaml")})
solarpanels = arrowdict({p.stem: p for p in SOLARPANEL_DIRECTORY.glob("*.yaml")})
cspinstallations = arrowdict(
    {p.stem: p for p in CSPINSTALLATION_DIRECTORY.glob("*.yaml")}
)



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT

import sys

import numpy as np
import xarray as xr
from dask.array import arccos, arcsin, arctan, cos, logical_and, radians, sin
from numpy import pi


def get_orientation(name, **params):
    """
    Definitions:
    -`slope` is the angle between ground and panel.
    -`azimuth` is the clockwise angle from North.
        i.e. azimuth = 180 faces exactly South
    """
    if isinstance(name, dict):
        params = name
        name = params.pop("name", "constant")
    return getattr(sys.modules[__name__], f"make_{name}")(**params)


def make_latitude_optimal():
    """
    Returns an optimal tilt angle for the given ``lat``, assuming that the
    panel is facing towards the equator, using a simple method from [1].

    This method only works for latitudes between 0 and 50. For higher
    latitudes, a static 40 degree angle is returned.

    These results should be used with caution, but there is some
    evidence that tilt angle may not be that important [2].

    Function and documentation has been adapted from gsee [3].

    [1] http://www.solarpaneltilt.com/#fixed
    [2] http://dx.doi.org/10.1016/j.solener.2010.12.014
    [3] https://github.com/renewables-ninja/gsee/blob/master/gsee/pv.py

    Parameters
    ----------
    lat : float
        Latitude in degrees.

    """

    def latitude_optimal(lon, lat, solar_position):
        slope = np.empty_like(lat.values)

        below_25 = np.abs(lat.values) <= np.radians(25)
        below_50 = np.abs(lat.values) <= np.radians(50)

        slope[below_25] = 0.87 * np.abs(lat.values[below_25])
        slope[~below_25 & below_50] = 0.76 * np.abs(
            lat.values[~below_25 & below_50]
        ) + np.radians(0.31)
        slope[~below_50] = np.radians(40.0)

        # South orientation for panels on northern hemisphere and vice versa
        azimuth = np.where(lat.values < 0, 0, pi)
        return dict(
            slope=xr.DataArray(slope, coords=lat.coords),
            azimuth=xr.DataArray(azimuth, coords=lat.coords),
        )

    return latitude_optimal


def make_constant(slope, azimuth):
    slope = radians(slope)
    azimuth = radians(azimuth)

    def constant(lon, lat, solar_position):
        return dict(slope=slope, azimuth=azimuth)

    return constant


def make_latitude(azimuth=180):
    azimuth = radians(azimuth)

    def latitude(lon, lat, solar_position):
        return dict(slope=lat, azimuth=azimuth)

    return latitude


def SurfaceOrientation(ds, solar_position, orientation, tracking=None):
    """
    Compute cos(incidence) for slope and panel azimuth.

    References
    ----------
    [1] Sproul, A. B., Derivation of the solar geometric relationships using
    vector analysis, Renewable Energy, 32(7), 1187–1205 (2007).
    [2] Marion, William F., and Aron P. Dobos. Rotation angle for the optimum
    tracking of one-axis trackers. No. NREL/TP-6A20-58891. National Renewable
    Energy Lab.(NREL), Golden, CO (United States), 2013.

    """
    lon = radians(ds["lon"])
    lat = radians(ds["lat"])

    orientation = orientation(lon, lat, solar_position)
    surface_slope = orientation["slope"]
    surface_azimuth = orientation["azimuth"]

    sun_altitude = solar_position["altitude"]
    sun_azimuth = solar_position["azimuth"]

    if tracking is None:
        cosincidence = sin(surface_slope) * cos(sun_altitude) * cos(
            surface_azimuth - sun_azimuth
        ) + cos(surface_slope) * sin(sun_altitude)

    elif tracking == "horizontal":  # horizontal tracking with horizontal axis
        axis_azimuth = orientation[
            "azimuth"
        ]  # here orientation['azimuth'] refers to the azimuth of the tracker axis.
        rotation = arctan(
            (cos(sun_altitude) / sin(sun_altitude)) * sin(sun_azimuth - axis_azimuth)
        )
        surface_slope = abs(rotation)
        surface_azimuth = axis_azimuth + arcsin(
            sin(rotation / sin(surface_slope))
        )  # the 2nd part yields +/-1 and determines if the panel is facing east or west
        cosincidence = cos(surface_slope) * sin(sun_altitude) + sin(
            surface_slope
        ) * cos(sun_altitude) * cos(sun_azimuth - surface_azimuth)

    elif tracking == "tilted_horizontal":  # horizontal tracking with tilted axis'
        axis_tilt = orientation[
            "slope"
        ]  # here orientation['slope'] refers to the tilt of the tracker axis.

        rotation = arctan(
            (cos(sun_altitude) * sin(sun_azimuth - surface_azimuth))
            / (
                cos(sun_altitude) * cos(sun_azimuth - surface_azimuth) * sin(axis_tilt)
                + sin(sun_altitude) * cos(axis_tilt)
            )
        )

        surface_slope = arccos(cos(rotation) * cos(axis_tilt))

        azimuth_difference = sun_azimuth - surface_azimuth
        azimuth_difference = np.where(
            azimuth_difference > pi, 2 * pi - azimuth_difference, azimuth_difference
        )
        azimuth_difference = np.where(
            azimuth_difference < -pi, 2 * pi + azimuth_difference, azimuth_difference
        )
        rotation = np.where(
            logical_and(rotation < 0, azimuth_difference > 0),
            rotation + pi,
            rotation,
        )
        rotation = np.where(
            logical_and(rotation > 0, azimuth_difference < 0),
            rotation - pi,
            rotation,
        )

        cosincidence = cos(rotation) * (
            sin(axis_tilt) * cos(sun_altitude) * cos(sun_azimuth - surface_azimuth)
            + cos(axis_tilt) * sin(sun_altitude)
        ) + sin(rotation) * cos(sun_altitude) * sin(sun_azimuth - surface_azimuth)

    elif tracking == "vertical":  # vertical tracking, surface azimuth = sun_azimuth
        cosincidence = sin(surface_slope) * cos(sun_altitude) + cos(
            surface_slope
        ) * sin(sun_altitude)
    elif tracking == "dual":  # both vertical and horizontal tracking
        cosincidence = np.float64(1.0)
    else:
        assert False, (
            "Values describing tracking system must be None for no tracking,"
            + "'horizontal' for 1-axis horizontal tracking,"
            + "tilted_horizontal' for 1-axis horizontal tracking of tilted panle,"
            + "vertical' for 1-axis vertical tracking, or 'dual' for 2-axis tracking"
        )

    # fixup incidence angle: if the panel is badly oriented and the sun shines
    # on the back of the panel (incidence angle > 90degree), the irradiation
    # would be negative instead of 0; this is prevented here.
    cosincidence = cosincidence.clip(min=0)

    return xr.Dataset(
        {
            "cosincidence": cosincidence,
            "slope": surface_slope,
            "azimuth": surface_azimuth,
        }
    )



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Module for loading gebco data.
"""

import logging

import rasterio as rio
import xarray as xr
from pandas import to_numeric
from rasterio.warp import Resampling

logger = logging.getLogger(__name__)

crs = 4326
features = {"height": ["height"]}


def get_data_gebco_height(xs, ys, gebco_path):
    x, X = xs.data[[0, -1]]
    y, Y = ys.data[[0, -1]]

    dx = (X - x) / (len(xs) - 1)
    dy = (Y - y) / (len(ys) - 1)

    with rio.open(gebco_path) as dataset:
        window = dataset.window(x - dx / 2, y - dy / 2, X + dx / 2, Y + dy / 2)
        gebco = dataset.read(
            indexes=1,
            window=window,
            out_shape=(len(ys), len(xs)),
            resampling=Resampling.average,
        )
        gebco = gebco[::-1]  # change inversed y-axis
        tags = dataset.tags(bidx=1)
        tags = {k: to_numeric(v, errors="ignore") for k, v in tags.items()}

    return xr.DataArray(
        gebco, coords=[("y", ys.data), ("x", xs.data)], name="height", attrs=tags
    )


def get_data(
    cutout,
    feature,
    tmpdir,
    monthly_requests=False,
    concurrent_requests=False,
    **creation_parameters,
):
    """
    Get the gebco height data.

    Parameters
    ----------
    cutout : atlite.Cutout
    feature : str
        Takes no effect, only here for consistency with other dataset modules.
    tmpdir : str
        Takes no effect, only here for consistency with other dataset modules.
    monthly_requests : bool
        Takes no effect, only here for consistency with other dataset modules.
    concurrent_requests : bool
        Takes no effect, only here for consistency with other dataset modules.
    **creation_parameters :
        Must include `gebco_path`.

    Returns
    -------
    xr.Dataset

    """
    if "gebco_path" not in creation_parameters:
        logger.error('Argument "gebco_path" not defined')
    path = creation_parameters["gebco_path"]

    coords = cutout.coords
    # assign time dimension even if not used
    return (
        get_data_gebco_height(coords["x"], coords["y"], path)
        .to_dataset()
        .assign_coords(cutout.coords)
    )



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Created on Mon May 11 11:15:41 2020.

@author: fabian
"""

import os
import sys
from datetime import date

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import urllib3
from dateutil.relativedelta import relativedelta
from shapely.geometry import LineString as Line
from shapely.geometry import Point

import atlite
from atlite import Cutout

urllib3.disable_warnings()

# %% Predefine tests for cutout

TIME = "2013-01-01"
BOUNDS = (-4, 56, 1.5, 62)
SARAH_DIR = os.getenv("SARAH_DIR", "/home/vres/climate-data/sarah_v2")
GEBCO_PATH = os.getenv("GEBCO_PATH", "/home/vres/climate-data/GEBCO_2014_2D.nc")


def all_notnull_test(cutout):
    """
    Test if no nan's in the prepared data occur.
    """
    assert cutout.data.notnull().all()


def prepared_features_test(cutout):
    """
    The prepared features series should contain all variables in cutout.data.
    """
    assert set(cutout.prepared_features) == set(cutout.data)


def merge_test(cutout, other, target_modules):
    merge = cutout.merge(other, compat="override")
    assert set(merge.module) == set(target_modules)


def wrong_recreation(cutout):
    with pytest.warns(UserWarning):
        Cutout(path=cutout.path, module="somethingelse")


def pv_test(cutout, time=TIME, skip_optimal_sum_test=False):
    """
    Test the atlite.Cutout.pv function with different settings.

    Compare optimal orientation with flat orientation.
    """
    orientation = {"slope": 0.0, "azimuth": 0.0}
    cap_factor = cutout.pv(atlite.resource.solarpanels.CdTe, orientation)

    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0

    production = cutout.pv(
        atlite.resource.solarpanels.CdTe, orientation, layout=cap_factor
    )

    assert production.notnull().all()
    assert production.sel(time=time + " 00:00") == 0

    cells = cutout.grid
    cells = cells.assign(regions=["lower"] * 200 + ["upper"] * (len(cells) - 200))
    shapes = cells.dissolve("regions")
    production, capacity = cutout.pv(
        atlite.resource.solarpanels.CdTe,
        orientation,
        layout=cap_factor,
        shapes=shapes,
        return_capacity=True,
    )
    cap_per_region = (
        cells.assign(cap_factor=cap_factor.stack(spatial=["y", "x"]).values)
        .groupby("regions")
        .cap_factor.sum()
    )

    assert all(cap_per_region.round(3) == capacity.round(3))

    # Now compare with optimal orienation
    cap_factor_opt = cutout.pv(atlite.resource.solarpanels.CdTe, "latitude_optimal")

    if not skip_optimal_sum_test:
        assert cap_factor_opt.sum() > cap_factor.sum()

    production_opt = cutout.pv(
        atlite.resource.solarpanels.CdTe, "latitude_optimal", layout=cap_factor_opt
    )

    assert production_opt.sel(time=time + " 00:00") == 0

    if not skip_optimal_sum_test:
        assert production_opt.sum() > production.sum()

    # now use the non simple trigon model
    production_other = cutout.pv(
        atlite.resource.solarpanels.CdTe,
        "latitude_optimal",
        layout=cap_factor_opt,
        trigon_model="other",
    )

    assert production_other.sel(time=time + " 00:00") == 0
    # should be roughly the same
    assert (production_other.sum() / production_opt.sum()).round(0) == 1

    # now another solarpanel with bofinger model
    production_other = cutout.pv(
        atlite.resource.solarpanels.KANENA,
        "latitude_optimal",
        layout=cap_factor_opt,
    )

    assert production_other.sel(time=time + " 00:00") == 0
    # should be roughly the same
    assert (production_other.sum() / production_opt.sum()).round(0) == 1


def pv_tracking_test(cutout):
    """
    Test the atlite.Cutout.pv function with different tracking settings and
    compare results.
    """
    orientation = {"slope": 0.0, "azimuth": 180.0}
    # tracking = None is the default option
    cap_factor = cutout.pv(
        atlite.resource.solarpanels.CSi,
        orientation,
        capacity_factor=True,
    )
    cap_factor_tracking_0axis = cutout.pv(
        atlite.resource.solarpanels.CSi,
        orientation,
        tracking=None,
        capacity_factor=True,
    )

    assert cap_factor_tracking_0axis.notnull().all()
    assert cap_factor_tracking_0axis.sum() > 0
    assert cap_factor.mean() == cap_factor_tracking_0axis.mean()

    # calculate with tracking = 'horizontal', 'tilted_horizontal', 'vertical' and 'dual', and compare tracking option results
    cap_factor_tracking_1axis_h = cutout.pv(
        atlite.resource.solarpanels.CSi,
        orientation,
        tracking="horizontal",
        capacity_factor=True,
    )

    cap_factor_tracking_1axis_th = cutout.pv(
        atlite.resource.solarpanels.CSi,
        orientation,
        tracking="tilted_horizontal",
        capacity_factor=True,
    )

    cap_factor_tracking_1axis_v = cutout.pv(
        atlite.resource.solarpanels.CSi,
        orientation,
        tracking="vertical",
        capacity_factor=True,
    )

    cap_factor_tracking_2axis = cutout.pv(
        atlite.resource.solarpanels.CSi,
        orientation,
        tracking="dual",
        capacity_factor=True,
    )

    assert cap_factor_tracking_1axis_v.notnull().all()
    assert cap_factor_tracking_1axis_v.sum() > 0
    assert cap_factor_tracking_1axis_v.mean() >= cap_factor_tracking_0axis.mean()

    assert cap_factor_tracking_1axis_h.notnull().all()
    assert cap_factor_tracking_1axis_h.sum() > 0
    assert cap_factor_tracking_1axis_h.mean() >= cap_factor_tracking_0axis.mean()

    assert cap_factor_tracking_1axis_th.notnull().all()
    assert cap_factor_tracking_1axis_th.sum() > 0
    assert cap_factor_tracking_1axis_th.mean() >= cap_factor_tracking_0axis.mean()

    assert cap_factor_tracking_2axis.notnull().all()
    assert cap_factor_tracking_2axis.sum() > 0
    assert cap_factor_tracking_2axis.mean() >= cap_factor_tracking_1axis_v.mean()
    assert cap_factor_tracking_2axis.mean() >= cap_factor_tracking_1axis_h.mean()
    assert cap_factor_tracking_2axis.mean() >= cap_factor_tracking_1axis_th.mean()


def csp_test(cutout):
    """
    Test the atlite.Cutout.csp function with different for different settings
    and technologies.
    """
    ## Test technology = "solar tower"
    st = cutout.csp(atlite.cspinstallations.SAM_solar_tower, capacity_factor=True)

    assert st.notnull().all()
    assert (st >= 0).all()
    assert (st <= 1).all()

    # Efficiencies <= 1 should lead to the conversion to always be less than perfect
    st = cutout.csp(atlite.cspinstallations.SAM_solar_tower)
    ll = cutout.csp(atlite.cspinstallations.lossless_installation)
    assert (st <= ll).all()

    ## Test technology = "parabolic trough"
    pt = cutout.csp(atlite.cspinstallations.SAM_parabolic_trough, capacity_factor=True)

    assert pt.notnull().all()
    assert (pt >= 0).all()
    assert (pt <= 1).all()

    # Efficiencies <= 1 should lead to the conversion to always be less than perfect
    pt = cutout.csp(atlite.cspinstallations.SAM_parabolic_trough)
    ll = cutout.csp(atlite.cspinstallations.lossless_installation)
    assert (pt <= ll).all()


def solar_thermal_test(cutout):
    """
    Test the atlite.Cutout.solar_thermal function with different settings.
    """
    cap_factor = cutout.solar_thermal()
    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0


def heat_demand_test(cutout):
    """
    Test the atlite.Cutout.heat_demand function with different settings.
    """
    demand = cutout.heat_demand()
    assert demand.notnull().all()
    assert demand.sum() > 0


def soil_temperature_test(cutout):
    """
    Test the atlite.Cutout.soil_temperature function with different settings.
    """
    demand = cutout.soil_temperature()
    assert demand.notnull().all()
    assert demand.sum() > 0


def dewpoint_temperature_test(cutout):
    """
    Test the atlite.Cutout.dewpoint_temperature function with different
    settings.
    """
    demand = cutout.dewpoint_temperature()
    assert demand.notnull().all()
    assert demand.sum() > 0


def wind_test(cutout):
    """
    Test the atlite.Cutout.wind function with two different layouts.

    The standard layout proportional to the capacity factors must have a
    lower production than a layout proportionally to the capacity layout
    squared.
    """
    cap_factor = cutout.wind(atlite.windturbines.Enercon_E101_3000kW)

    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0

    production = cutout.wind(atlite.windturbines.Enercon_E101_3000kW, layout=cap_factor)

    assert production.notnull().all()
    assert production.sum() > 0

    # Now create a better layout with same amount of installed power
    better_layout = (cap_factor**2) / (cap_factor**2).sum() * cap_factor.sum()
    better_production = cutout.wind(
        atlite.windturbines.Enercon_E101_3000kW, layout=better_layout
    )

    assert better_production.sum() > production.sum()

    # now use smooth wind power curve
    production = cutout.wind(
        atlite.windturbines.Enercon_E101_3000kW, layout=cap_factor, smooth=True
    )

    assert production.notnull().all()
    assert production.sum() > 0

    # test with different power law interpolation method
    production = cutout.wind(
        atlite.windturbines.Enercon_E101_3000kW,
        layout=cap_factor,
        interpolation_method="power",
    )


def runoff_test(cutout):
    """
    Test the atlite.Cutout.runoff function.

    First check if the total of all capacity factors is not null. Then
    compare the runoff at sites which belong to the lower (altitude)
    half of the map, with the runoff at higher sites. The runoff at the
    lower sites (mostly at sea level) should have a smaller capacity
    factor total and production.
    """
    cap_factor = cutout.runoff()
    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0

    height = cutout.data.height.load()
    q = np.quantile(height, 0.5)
    lower_area = height <= q
    higher_area = height > q
    assert cap_factor.where(lower_area).sum() < cap_factor.where(higher_area).sum()

    low_level_prod = cutout.runoff(layout=cap_factor.where(lower_area, 0))
    high_level_prod = cutout.runoff(layout=cap_factor.where(higher_area, 0))
    assert low_level_prod.sum() < high_level_prod.sum()


def hydro_test(cutout):
    """
    Test the atlite.Cutout.hydro function.
    """
    plants = pd.DataFrame(
        cutout.grid.loc[[0], ["x", "y"]].values, columns=["lon", "lat"]
    )
    basins = gpd.GeoDataFrame(
        dict(
            geometry=[cutout.grid.geometry[0]],
            HYBAS_ID=[0],
            DIST_MAIN=10,
            NEXT_DOWN=None,
        ),
        index=[0],
        crs=cutout.crs,
    )
    ds = cutout.hydro(plants, basins)
    assert ds.sel(plant=0).sum() > 0


def line_rating_test(cutout):
    shapes = [Line([Point(-3, 57), Point(0, 60)])]
    resistance = 0.06 * 1e-3
    i = cutout.line_rating(shapes, resistance)
    assert i.notnull().all().item()


def coefficient_of_performance_test(cutout):
    """
    Test the coefficient_of_performance function.
    """
    cap_factor = cutout.coefficient_of_performance(source="air")
    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0

    cap_factor = cutout.coefficient_of_performance(source="soil")
    assert cap_factor.notnull().all()
    assert cap_factor.sum() > 0


class TestERA5:
    @staticmethod
    def test_data_module_arguments_era5(cutout_era5):
        """
        All data variables should have an attribute to which module they
        belong.
        """
        for v in cutout_era5.data:
            assert cutout_era5.data.attrs["module"] == "era5"

    @staticmethod
    def test_all_non_na_era5(cutout_era5):
        """
        Every cells should have data.
        """
        assert np.isfinite(cutout_era5.data).all()

    @staticmethod
    def test_all_non_na_era5_coarse(cutout_era5_coarse):
        """
        Every cells should have data.
        """
        assert np.isfinite(cutout_era5_coarse.data).all()

    @staticmethod
    @pytest.mark.skipif(
        os.name == "nt",
        reason="This test breaks on windows machine on CI" " due to unknown reasons.",
    )
    def test_all_non_na_era5_weird_resolution(cutout_era5_weird_resolution):
        """
        Every cells should have data.
        """
        assert np.isfinite(cutout_era5_weird_resolution.data).all()

    @staticmethod
    def test_dx_dy_preservation_era5(cutout_era5):
        """
        The coordinates should be the same after preparation.
        """
        assert np.allclose(np.diff(cutout_era5.data.x), 0.25)
        assert np.allclose(np.diff(cutout_era5.data.y), 0.25)

    @staticmethod
    def test_dx_dy_preservation_era5_coarse(cutout_era5_coarse):
        """
        The coordinates should be the same after preparation.
        """
        assert np.allclose(
            np.diff(cutout_era5_coarse.data.x), cutout_era5_coarse.data.attrs["dx"]
        )
        assert np.allclose(
            np.diff(cutout_era5_coarse.data.y), cutout_era5_coarse.data.attrs["dy"]
        )

    @staticmethod
    @pytest.mark.skipif(
        os.name == "nt",
        reason="This test breaks on windows machine on CI" " due to unknown reasons.",
    )
    def test_dx_dy_preservation_era5_weird_resolution(cutout_era5_weird_resolution):
        """
        The coordinates should be the same after preparation.
        """
        assert np.allclose(
            np.diff(cutout_era5_weird_resolution.data.x),
            cutout_era5_weird_resolution.data.attrs["dx"],
        )
        assert np.allclose(
            np.diff(cutout_era5_weird_resolution.data.y),
            cutout_era5_weird_resolution.data.attrs["dy"],
        )

    @staticmethod
    def test_compare_with_get_data_era5(cutout_era5, tmp_path):
        """
        The prepared data should be exactly the same as from the low level
        function.
        """
        # TODO Needs fix
        pass
        # influx = atlite.datasets.era5.get_data(cutout_era5, "influx", tmpdir=tmp_path)
        # assert_allclose(
        #     influx.influx_toa, cutout_era5.data.influx_toa, atol=1e-5, rtol=1e-5
        # )

    @staticmethod
    def test_prepared_features_era5(cutout_era5):
        return prepared_features_test(cutout_era5)

    @staticmethod
    @pytest.mark.skipif(
        sys.platform == "win32", reason="NetCDF update not working on windows"
    )
    @staticmethod
    def test_wrong_loading(cutout_era5):
        wrong_recreation(cutout_era5)

    @staticmethod
    def test_pv_era5(cutout_era5):
        return pv_test(cutout_era5)

    @staticmethod
    def test_pv_tracking_era5(cutout_era5):
        return pv_tracking_test(cutout_era5)

    @staticmethod
    def test_pv_era5_2days_crossing_months(cutout_era5_2days_crossing_months):
        """
        See https://github.com/PyPSA/atlite/issues/256.
        """
        return pv_test(cutout_era5_2days_crossing_months, time="2013-03-01")

    @staticmethod
    def test_pv_era5_3h_sampling(cutout_era5_3h_sampling):
        assert pd.infer_freq(cutout_era5_3h_sampling.data.time) == "3h"
        return pv_test(cutout_era5_3h_sampling)

    @staticmethod
    def test_pv_era5_and_era5t(cutout_era5t):
        """
        CDSAPI returns ERA5T data for the *previous* month, and ERA5 data for
        the *second-previous* month. We request data spanning 2 days between the 2
        months to test merging ERA5 data with ERA5T.

        See documentation here: https://confluence.ecmwf.int/pages/viewpage.action?pageId=173385064

        Note: the above page says that ERA5 data are made available with a *3* month delay,
        but experience shows that it's with a *2* month delay. Hence the test with previous
        vs. second-previous month.
        """
        today = date.today()
        first_day_this_month = today.replace(day=1)
        first_day_prev_month = first_day_this_month - relativedelta(months=1)
        last_day_second_prev_month = first_day_prev_month - relativedelta(days=1)

        # If ERA5 and ERA5T data are merged successfully, there should be no null values
        # in any of the features of the cutout
        for feature in cutout_era5t.data.values():
            assert feature.notnull().to_numpy().all()

        # temporarily skip the optimal sum test, as there seems to be a bug in the
        # optimal orientation calculation. See https://github.com/PyPSA/atlite/issues/358
        pv_test(
            cutout_era5t,
            time=str(last_day_second_prev_month),
            skip_optimal_sum_test=True,
        )
        return pv_test(
            cutout_era5t, time=str(first_day_prev_month), skip_optimal_sum_test=True
        )

    @staticmethod
    def test_wind_era5(cutout_era5):
        return wind_test(cutout_era5)

    @staticmethod
    def test_runoff_era5(cutout_era5):
        return runoff_test(cutout_era5)

    @staticmethod
    def test_hydro_era5(cutout_era5):
        return hydro_test(cutout_era5)

    @staticmethod
    def test_solar_thermal_era5(cutout_era5):
        return solar_thermal_test(cutout_era5)

    @staticmethod
    def test_heat_demand_era5(cutout_era5):
        return heat_demand_test(cutout_era5)

    @staticmethod
    def test_soil_temperature_era5(cutout_era5):
        return soil_temperature_test(cutout_era5)

    @staticmethod
    def test_dewpoint_temperature_era5(cutout_era5):
        return dewpoint_temperature_test(cutout_era5)

    @staticmethod
    def test_line_rating_era5(cutout_era5):
        return line_rating_test(cutout_era5)


@pytest.mark.skipif(
    not os.path.exists(SARAH_DIR), reason="'sarah_dir' is not a valid path"
)
class TestSarah:
    @staticmethod
    def test_all_non_na_sarah(cutout_sarah):
        """
        Every cells should have data.
        """
        assert np.isfinite(cutout_sarah.data).all()

    @staticmethod
    def test_all_non_na_sarah_fine(cutout_sarah_fine):
        """
        Every cells should have data.
        """
        assert np.isfinite(cutout_sarah_fine.data).all()

    @staticmethod
    def test_all_non_na_sarah_weird_resolution(cutout_sarah_weird_resolution):
        """
        Every cells should have data.
        """
        assert np.isfinite(cutout_sarah_weird_resolution.data).all()

    @staticmethod
    def test_dx_dy_preservation_sarah(cutout_sarah):
        """
        The coordinates should be the same after preparation.
        """
        assert np.allclose(np.diff(cutout_sarah.data.x), 0.25)
        assert np.allclose(np.diff(cutout_sarah.data.y), 0.25)

    @staticmethod
    def test_prepared_features_sarah(cutout_sarah):
        return prepared_features_test(cutout_sarah)

    @staticmethod
    def test_merge(cutout_sarah, cutout_era5):
        return merge_test(cutout_sarah, cutout_era5, ["sarah", "era5"])

    @staticmethod
    def test_pv_sarah(cutout_sarah):
        return pv_test(cutout_sarah)

    @staticmethod
    def test_wind_sarah(cutout_sarah):
        return wind_test(cutout_sarah)

    @staticmethod
    def test_runoff_sarah(cutout_sarah):
        return runoff_test(cutout_sarah)


@pytest.mark.skipif(
    not os.path.exists(GEBCO_PATH), reason="'gebco_path' is not a valid path"
)
class TestGebco:
    @staticmethod
    def test_all_non_na_gebco(cutout_gebco):
        """
        Every cells should have data.
        """
        assert np.isfinite(cutout_gebco.data).all()



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT

import numpy as np

# Huld model was copied from gsee -- global solar energy estimator
# by Stefan Pfenninger
# https://github.com/renewables-ninja/gsee/blob/master/gsee/pv.py


def _power_huld(irradiance, t_amb, pc):
    """
    AC power per capacity predicted by Huld model, based on W/m2 irradiance.

    Maximum power point tracking is assumed.

    [1] Huld, T. et al., 2010. Mapping the performance of PV modules,
    effects of module type and data averaging. Solar Energy, 84(2),
    p.324-338. DOI: 10.1016/j.solener.2009.12.002
    """
    # normalized module temperature
    T_ = (pc["c_temp_amb"] * t_amb + pc["c_temp_irrad"] * irradiance) - pc["r_tmod"]

    # normalized irradiance
    G_ = irradiance / pc["r_irradiance"]

    log_G_ = np.log(G_.where(G_ > 0))
    # NB: np.log without base implies base e or ln
    eff = (
        1
        + pc["k_1"] * log_G_
        + pc["k_2"] * (log_G_) ** 2
        + T_ * (pc["k_3"] + pc["k_4"] * log_G_ + pc["k_5"] * log_G_**2)
        + pc["k_6"] * (T_**2)
    )

    eff = eff.fillna(0.0).clip(min=0)

    da = G_ * eff * pc.get("inverter_efficiency", 1.0)
    da.attrs["units"] = "kWh/kWp"
    da = da.rename("specific generation")

    return da


def _power_bofinger(irradiance, t_amb, pc):
    """
    AC power per capacity predicted by bofinger model, based on W/m2
    irradiance.

    Maximum power point tracking is assumed.

    [2] Hans Beyer, Gerd Heilscher and Stefan Bofinger, 2004. A robust
    model for the MPP performance of different types of PV-modules
    applied for the performance check of grid connected systems.
    """
    fraction = (pc["NOCT"] - pc["Tamb"]) / pc["Intc"]

    eta_ref = (
        pc["A"]
        + pc["B"] * irradiance
        + pc["C"] * np.log(irradiance.where(irradiance != 0))
    )
    eta = (
        eta_ref
        * (1.0 + pc["D"] * (fraction * irradiance + (t_amb - pc["Tstd"])))
        / (1.0 + pc["D"] * fraction / pc["ta"] * eta_ref * irradiance)
    ).fillna(0)

    capacity = (pc["A"] + pc["B"] * 1000.0 + pc["C"] * np.log(1000.0)) * 1e3
    power = irradiance * eta * (pc.get("inverter_efficiency", 1.0) / capacity)
    power = power.where(irradiance >= pc["threshold"], 0)
    return power.rename("AC power")


def SolarPanelModel(ds, irradiance, pc):
    model = pc.get("model", "huld")

    if model == "huld":
        return _power_huld(irradiance, ds["temperature"], pc)
    elif model == "bofinger":
        return _power_bofinger(irradiance, ds["temperature"], pc)
    else:
        AssertionError(f"Unknown panel model: {model}")



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Base class for atlite.
"""

# There is a binary incompatibility between the pip wheels of netCDF4 and
# rasterio, which leads to the first one to work correctly while the second
# loaded one fails by loading netCDF4 first, we ensure that most of atlite's
# functionality works fine, even when the pip wheels have been used, only for
# resampling the sarah dataset it is important to use conda.
# Refer to
# https://github.com/pydata/xarray/issues/2535,
# https://github.com/rasterio/rasterio-wheels/issues/12

import logging
from pathlib import Path
from tempfile import mktemp
from warnings import warn

import dask
import geopandas as gpd
import numpy as np
import pandas as pd
import rasterio as rio
import xarray as xr
from numpy import append, atleast_1d
from pyproj import CRS
from shapely.geometry import box

from atlite.convert import (
    coefficient_of_performance,
    convert_and_aggregate,
    csp,
    dewpoint_temperature,
    heat_demand,
    hydro,
    irradiation,
    line_rating,
    pv,
    runoff,
    soil_temperature,
    solar_thermal,
    temperature,
    wind,
)
from atlite.data import available_features, cutout_prepare
from atlite.datasets import modules as datamodules
from atlite.gis import (
    compute_availabilitymatrix,
    compute_indicatormatrix,
    compute_intersectionmatrix,
    get_coords,
)
from atlite.utils import CachedAttribute

logger = logging.getLogger(__name__)


class Cutout:
    """
    Cutout base class.

    This class builds the starting point for most atlite
    functionalities.
    """

    def __init__(self, path, **cutoutparams):
        """
        Provide an atlite cutout object.

        Create a cutout object to use atlite operations on it. Based on the
        provided parameters, atlite first checks whether this cutout already
        exists on disk and if yes, loads this cutout.

        If the cutout does not yet exist on disk, then atlite creates an
        "unprepared" cutout object. This does not yet contain the full data.
        The process of preparing (loading the data) can then be started with
        `cutout.prepare()`.

        Parameters
        ----------
        path : str | path-like
            NetCDF from which to load or where to store the cutout.
        module : str or list
            The dataset(s) which works as a basis for the cutout. Available
            modules are "era5", "sarah" and "gebco".
            This is necessary when building a new cutout.
            If more than one module is given, their order determines how atlite
            fills up missing features when preparing the cutout with
            `Cutout.prepare()`. For example `influx_diffuse` is provided by
            the `sarah` and the `era5` module. Prioritizing sarah and setting
            module=['sarah', 'era5'] will load `influx_diffuse` from the sarah
            module and ignoring the era5 'influx_diffuse' data.
        time : str | slice
            Time range to include in the cutout, e.g. "2011" or
            ("2011-01-05", "2011-01-25")
            This is necessary when building a new cutout.
        bounds : GeoSeries.bounds | DataFrame, optional
            The outer bounds of the cutout or as a DataFrame
            containing (min.long, min.lat, max.long, max.lat).
        x : slice, optional
            Outer longitudinal bounds for the cutout (west, east).
        y : slice, optional
            Outer latitudinal bounds for the cutout (south, north).
        dx : float, optional
            Step size of the x coordinate. The default is 0.25.
        dy : float, optional
            Step size of the y coordinate. The default is 0.25.
        dt : str, optional
            Frequency of the time coordinate. The default is 'h'. Valid are all
            pandas offset aliases.
        chunks : dict
            Chunks when opening netcdf files. For cutout preparation recommand
            to chunk only along the time dimension. Defaults to {'time': 20}
        data : xr.Dataset
            User provided cutout data. Save the cutout using `Cutout.to_file()`
            afterwards.

        Other Parameters
        ----------------
        sanitize : bool, default True
            Whether to sanitize the data when preparing the cutout. Takes
            effect for 'era5' data loading.
        sarah_dir : str, Path
            Directory of on-disk sarah data. This must be given when using the
            sarah module.
        sarah_interpolate : bool, default True
            Whether to interpolate NaN's in the SARAH data. This takes effect for
            sarah data which has missing data for areas where dawn and
            nightfall happens (ca. 30 min gap).
        gebco_path: str
            Path to find the gebco netcdf file. Only necessary when including
            the gebco module.
        parallel : bool, default False
            Whether to open dataset in parallel mode. Take effect for all
            xr.open_mfdataset usages.

        """
        path = Path(path).with_suffix(".nc")
        chunks = cutoutparams.pop("chunks", {"time": 100})
        if isinstance(chunks, dict):
            storable_chunks = {f"chunksize_{k}": v for k, v in (chunks or {}).items()}
        else:
            storable_chunks = {}

        # Three cases. First, cutout exists -> take the data.
        # Second, data is given -> take it. Third, else -> build a new cutout
        if path.is_file():
            data = xr.open_dataset(str(path))
            data = data.chunk(chunks)
            data.attrs.update(storable_chunks)
            if cutoutparams:
                warn(
                    f'Arguments {", ".join(cutoutparams)} are ignored, since '
                    "cutout is already built."
                )
        elif "data" in cutoutparams:
            data = cutoutparams.pop("data")
        else:
            logger.info(f"Building new cutout {path}")

            if "bounds" in cutoutparams:
                x1, y1, x2, y2 = cutoutparams.pop("bounds")
                cutoutparams.update(x=slice(x1, x2), y=slice(y1, y2))

            try:
                x = cutoutparams.pop("x")
                y = cutoutparams.pop("y")
                time = cutoutparams.pop("time")
                module = cutoutparams.pop("module")
            except KeyError as exc:
                raise TypeError(
                    "Arguments 'time' and 'module' must be "
                    "specified. Spatial bounds must either be "
                    "passed via argument 'bounds' or 'x' and 'y'."
                ) from exc

            # TODO: check for dx, dy, x, y fine with module requirements
            coords = get_coords(x, y, time, **cutoutparams)

            attrs = {
                "module": module,
                "prepared_features": [],
                **storable_chunks,
                **cutoutparams,
            }
            data = xr.Dataset(coords=coords, attrs=attrs)

        # Check compatibility of CRS
        modules = atleast_1d(data.attrs.get("module"))
        crs = set(CRS(datamodules[m].crs) for m in modules)
        assert len(crs) == 1, f"CRS of {module} not compatible"

        self.path = path
        self.data = data

    @property
    def name(self):
        """
        Name of the cutout.
        """
        return self.path.stem

    @property
    def module(self):
        """
        Data module of the cutout.
        """
        return self.data.attrs.get("module")

    @property
    def crs(self):
        """
        Coordinate Reference System of the cutout.
        """
        return CRS(datamodules[atleast_1d(self.module)[0]].crs)

    @property
    def available_features(self):
        """
        List of available weather data features for the cutout.
        """
        return available_features(self.module)

    @property
    def chunks(self):
        """
        Chunking of the cutout data used by dask.
        """
        chunks = {
            k.lstrip("chunksize_"): v
            for k, v in self.data.attrs.items()
            if k.startswith("chunksize_")
        }
        return None if chunks == {} else chunks

    @property
    def coords(self):
        """
        Geographic coordinates of the cutout.
        """
        return self.data.coords

    @property
    def shape(self):
        """
        Size of spatial dimensions (y, x) of the cutout data.
        """
        return len(self.coords["y"]), len(self.coords["x"])

    @property
    def extent(self):
        """
        Total extent of the area covered by the cutout (x, X, y, Y).
        """
        xs, ys = self.coords["x"].values, self.coords["y"].values
        dx, dy = self.dx, self.dy
        return np.array(
            [xs[0] - dx / 2, xs[-1] + dx / 2, ys[0] - dy / 2, ys[-1] + dy / 2]
        )

    @property
    def bounds(self):
        """
        Total bounds of the area covered by the cutout (x, y, X, Y).
        """
        return self.extent[[0, 2, 1, 3]]

    @property
    def transform(self):
        """
        Get the affine transform of the cutout.
        """
        return rio.Affine(
            self.dx,
            0,
            self.coords["x"].values[0] - self.dx / 2,
            0,
            self.dy,
            self.coords["y"].values[0] - self.dy / 2,
        )

    @property
    def transform_r(self):
        """
        Get the affine transform of the cutout with reverse y-order.
        """
        return rio.Affine(
            self.dx,
            0,
            self.coords["x"].values[0] - self.dx / 2,
            0,
            -self.dy,
            self.coords["y"].values[-1] + self.dy / 2,
        )

    @property
    def dx(self):
        """
        Spatial resolution on the x coordinates.
        """
        x = self.coords["x"]
        return round((x[-1] - x[0]).item() / (x.size - 1), 8)

    @property
    def dy(self):
        """
        Spatial resolution on the y coordinates.
        """
        y = self.coords["y"]
        return round((y[-1] - y[0]).item() / (y.size - 1), 8)

    @property
    def dt(self):
        """
        Time resolution of the cutout.
        """
        return pd.infer_freq(self.coords["time"].to_index())

    @property
    def prepared(self):
        """
        Boolean indicating whether all available features are prepared.
        """
        return self.prepared_features.sort_index().equals(
            self.available_features.sort_index()
        )

    @property
    def prepared_features(self):
        """
        Get the list of prepared features in the cutout.
        """
        index = [
            (self.data[v].attrs["module"], self.data[v].attrs["feature"])
            for v in self.data
        ]
        index = pd.MultiIndex.from_tuples(index, names=["module", "feature"])
        return pd.Series(list(self.data), index, dtype=object)

    @CachedAttribute
    def grid(self):
        """
        Cutout grid with coordinates and geometries.

        The coordinates represent the centers of the grid cells.

        Returns
        -------
        geopandas.GeoDataFrame
            Frame with coordinate columns 'x' and 'y', and geometries of the
            corresponding grid cells.

        """
        xs, ys = np.meshgrid(self.coords["x"], self.coords["y"])
        coords = np.asarray((np.ravel(xs), np.ravel(ys))).T
        span = (coords[self.shape[1] + 1] - coords[0]) / 2
        cells = [box(*c) for c in np.hstack((coords - span, coords + span))]
        return gpd.GeoDataFrame(
            {"x": coords[:, 0], "y": coords[:, 1], "geometry": cells},
            crs=self.crs,
        )

    def sel(self, path=None, bounds=None, buffer=0, **kwargs):
        """
        Select parts of the cutout.

        Parameters
        ----------
        path : str | path-like
            File where to store the sub-cutout. Defaults to a temporary file.
        bounds : GeoSeries.bounds | DataFrame, optional
            The outer bounds of the cutout or as a DataFrame
            containing (min.long, min.lat, max.long, max.lat).
        buffer : float, optional
            Buffer around the bounds. The default is 0.
        **kwargs :
            Passed to `xr.Dataset.sel` for data selection.

        Returns
        -------
        selected : Cutout
            Selected cutout.

        """
        if path is None:
            path = mktemp(
                prefix=f"{self.path.stem}-",
                suffix=self.path.suffix,
                dir=self.path.parent,
            )

        if bounds is not None:
            if buffer > 0:
                bounds = box(*bounds).buffer(buffer).bounds
            x1, y1, x2, y2 = bounds
            kwargs.update(x=slice(x1, x2), y=slice(y1, y2))
        data = self.data.sel(**kwargs)
        return Cutout(path, data=data)

    def merge(self, other, path=None, **kwargs):
        """
        Merge two cutouts into a single cutout.

        Parameters
        ----------
        other : atlite.Cutout
            Other cutout to merge.
        path : str | path-like
            File where to store the merged cutout. Defaults to a temporary file.
        **kwargs
            Keyword arguments passed to `xarray.merge()`.

        Returns
        -------
        merged : Cutout
            Merged cutout.

        """
        assert isinstance(other, Cutout)

        if path is None:
            path = mktemp(
                prefix=f"{self.path.stem}-",
                suffix=self.path.suffix,
                dir=self.path.parent,
            )

        attrs = {**self.data.attrs, **other.data.attrs}
        attrs["module"] = list(set(append(*atleast_1d(self.module, other.module))))
        features = self.prepared_features.index.unique("feature")
        otherfeatures = other.prepared_features.index.unique("feature")
        attrs["prepared_features"] = list(features.union(otherfeatures))

        data = self.data.merge(other.data, **kwargs).assign_attrs(**attrs)

        return Cutout(path, data=data)

    def to_file(self, fn=None):
        """
        Save cutout to a netcdf file.

        Parameters
        ----------
        fn : str | path-like
            File name where to store the cutout, defaults to `cutout.path`.

        """
        if fn is None:
            fn = self.path
        self.data.to_netcdf(fn)

    def __repr__(self):
        start = np.datetime_as_string(self.coords["time"].values[0], unit="D")
        end = np.datetime_as_string(self.coords["time"].values[-1], unit="D")
        return (
            '<Cutout "{}">\n'
            " x = {:.2f} ⟷ {:.2f}, dx = {:.2f}\n"
            " y = {:.2f} ⟷ {:.2f}, dy = {:.2f}\n"
            " time = {} ⟷ {}, dt = {}\n"
            " module = {}\n"
            " prepared_features = {}".format(
                self.name,
                self.coords["x"].values[0],
                self.coords["x"].values[-1],
                self.dx,
                self.coords["y"].values[0],
                self.coords["y"].values[-1],
                self.dy,
                start,
                end,
                self.dt,
                self.module,
                list(self.prepared_features.index.unique("feature")),
            )
        )

    def indicatormatrix(self, shapes, shapes_crs=4326):
        """
        Compute the indicatormatrix.

        The indicatormatrix I[i,j] is a sparse representation of the ratio
        of the area in orig[j] lying in dest[i], where orig and dest are
        collections of polygons, i.e.

        A value of I[i,j] = 1 indicates that the shape orig[j] is fully
        contained in shape dest[j].

        Note that the polygons must be in the same crs.

        Parameters
        ----------
        shapes : Collection of shapely polygons

        Returns
        -------
        I : sp.sparse.lil_matrix
          Indicatormatrix

        """
        return compute_indicatormatrix(self.grid, shapes, self.crs, shapes_crs)

    def intersectionmatrix(self, shapes, shapes_crs=4326):
        """
        Compute the intersectionmatrix.

        The intersectionmatrix is a sparse matrix with entries (i,j) being one
        if shapes orig[j] and dest[i] are intersecting, and zero otherwise.

        Note that the polygons must be in the same crs.

        Parameters
        ----------
        orig : Collection of shapely polygons
        dest : Collection of shapely polygons

        Returns
        -------
        I : sp.sparse.lil_matrix
          Intersectionmatrix

        """
        return compute_intersectionmatrix(self.grid, shapes, self.crs, shapes_crs)

    def area(self, crs=None):
        """
        Get the area per grid cell as a DataArray with coords (x,y).

        Parameters
        ----------
        crs : int, optional
            The coordinate reference system (CRS) to use for the calculation.
            Defaults to the crs of the cutout.

        Returns
        -------
        xr.DataArray
            A DataArray containing the area per grid cell with coordinates (x,y).

        """
        if crs is None:
            crs = self.crs

        area = self.grid.to_crs(crs).area
        return xr.DataArray(
            area.values.reshape(self.shape),
            [self.coords["y"], self.coords["x"]],
        )

    def uniform_layout(self):
        """
        Get a uniform capacity layout for all grid cells.
        """
        return xr.DataArray(1, [self.coords["y"], self.coords["x"]])

    def uniform_density_layout(self, capacity_density, crs=None):
        """
        Get a capacity layout from a uniform capacity density.

        Parameters
        ----------
        capacity_density : float
            Capacity density in capacity/projection unit squared.
        crs : int, optional
            CRS to calculate the total area of grid cells.
            Passed to `Cutout.area()`.

        Returns
        -------
        xr.DataArray
            Capacity layout with dimensions 'x' and 'y' indicating the total
            capacity placed within one grid cell.

        """
        return capacity_density * self.area(crs)

    def equals(self, other):
        """
        It overrides xarray.Dataset.equals and ignores the path attribute in the comparison
        """
        if not isinstance(other, Cutout):
            return NotImplemented
        # Compare cutouts data attributes
        return self.data.equals(other.data)

    def layout_from_capacity_list(self, data, col="Capacity"):
        """
        Get a capacity layout aligned to the cutout based on a capacity list.

        Parameters
        ----------
        data : pandas.DataFrame
            Capacity list with columns 'x', 'y' and col. Each capacity entry
            is added to the grid cell intersecting with the coordinate (x,y).
        col : str, optional
            Name of the column with capacity values. The default is 'Capacity'.

        Returns
        -------
        xr.DataArray
            Capacity layout with dimensions 'x' and 'y' indicating the total
            capacity placed within one grid cell.

        Example
        -------
        >>> import atlite
        >>> import powerplantmatching as pm

        >>> data = pm.data.OPSD_VRE_country('DE')
        >>> data = (data.query('Fueltype == "Solar"')
                    .rename(columns={'lon':'x', 'lat':'y'}))

        >>> cutout = atlite.Cutout('Germany', x = slice(-5, 15), y = slice(40, 55),
                       time='2013-06-01', module='era5')
        >>> cutout.prepare(features=['influx', 'temperature'])
        >>> layout = cutout.layout_from_capacity_list(data)
        >>> pv = cutout.pv('CdTe', 'latitude_optimal', layout=layout)
        >>> pv.plot()

        """
        with dask.config.set(**{"array.slicing.split_large_chunks": False}):
            nearest = (
                self.uniform_layout()
                .chunk()
                .sel({"x": data.x.values, "y": data.y.values}, "nearest")
            )

        data = (
            data.assign(x=nearest.x.data, y=nearest.y.data)
            .groupby(["y", "x"])[col]
            .sum()
        )
        return data.to_xarray().reindex_like(self.data).fillna(0)

    availabilitymatrix = compute_availabilitymatrix

    # Preparation functions

    prepare = cutout_prepare

    # Conversion and aggregation functions

    convert_and_aggregate = convert_and_aggregate

    heat_demand = heat_demand

    temperature = temperature

    soil_temperature = soil_temperature

    dewpoint_temperature = dewpoint_temperature

    coefficient_of_performance = coefficient_of_performance

    solar_thermal = solar_thermal

    wind = wind

    irradiation = irradiation

    pv = pv

    csp = csp

    runoff = runoff

    hydro = hydro

    line_rating = line_rating



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Created on Wed May  6 15:23:13 2020.

@author: fabian
"""

# IDEAS for tests

import numpy as np
import pytest
import rasterio as rio
from xarray.testing import assert_equal

from atlite import Cutout

TIME = "2013-01-01"
X0 = -4.0
Y0 = 56.0
X1 = 1.5
Y1 = 61.0


@pytest.fixture
def ref():
    return Cutout(
        path="creation_ref", module="era5", bounds=(X0, Y0, X1, Y1), time=TIME
    )


def test_name(ref):
    assert ref.name == "creation_ref"


def test_module(ref):
    assert ref.module == "era5"


def test_crs(ref):
    assert ref.crs == "EPSG:4326"


def test_shape(ref):
    assert ref.shape == (21, 23)


def test_extent(ref):
    reference_extent = [-4.125, 1.625, 55.875, 61.125]
    assert all([x == y for x, y in zip(ref.extent, reference_extent)])


def test_bounds(ref):
    reference_extent = [-4.125, 55.875, 1.625, 61.125]
    assert all([x == y for x, y in zip(ref.bounds, reference_extent)])


def test_transform(ref):
    reference_affine = rio.Affine(
        0.25, 0.00, -4.125, 0.00, 0.25, 55.875, 0.00, 0.00, 1.00
    )
    assert reference_affine == ref.transform


def test_equals(ref):
    test_cutout = Cutout(
        path="creation_ref", module="era5", bounds=(X0, Y0, X1, Y1), time=TIME
    )
    assert ref.equals(test_cutout)


def test_transform_r(ref):
    reference_affine = rio.Affine(
        0.25, 0.00, -4.125, 0.00, -0.25, 61.125, 0.00, 0.00, 1.00
    )
    assert reference_affine == ref.transform_r


def test_odd_bounds_coords(ref):
    cutout = Cutout(
        path="odd_bounds",
        module="era5",
        time=TIME,
        bounds=(X0 - 0.1, Y0 - 0.02, X1 + 0.03, Y1 + 0.13),
    )
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())


def test_xy_coords(ref):
    cutout = Cutout(
        path="xy", module="era5", time=TIME, x=slice(X0, X1), y=slice(Y0, Y1)
    )
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())


def test_xy_reversed_coords(ref):
    cutout = Cutout(
        path="xy_r", module="era5", time=TIME, x=slice(X1, X0), y=slice(Y1, Y0)
    )
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())


def test_time_sclice_coords(ref):
    cutout = Cutout(
        path="time_slice",
        module="era5",
        time=slice("2013-01-01", "2013-01-01"),
        x=slice(X0, X1),
        y=slice(Y0, Y1),
    )
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())


def test_auto_chunking(ref):
    cutout = Cutout(
        path="auto_chunking",
        module="era5",
        time=slice("2013-01-01", "2013-01-01"),
        x=slice(X0, X1),
        y=slice(Y0, Y1),
        chunks="auto",
    )
    assert_equal(cutout.coords.to_dataset(), ref.coords.to_dataset())


def test_dx_dy_dt():
    """
    Test the properties dx, dy, dt of atlite.Cutout.

    The spatial resolution can be changed through the creation_params dx
    and dy, the time resolution is hard coded (deep in the modules...)
    to one hour.
    """
    dx = 0.5
    dy = 1
    cutout = Cutout(
        path="resolution",
        module="era5",
        time=slice("2013-01-01", "2013-01-01"),
        x=slice(X0, X1),
        y=slice(Y0, Y1),
        dx=dx,
        dy=dy,
    )
    assert dx == cutout.dx
    assert dy == cutout.dy
    assert "h" == cutout.dt


def test_available_features(ref):
    modules = ref.available_features.index.unique("module")
    assert len(modules) == 1
    assert modules[0] == "era5"

    cutout = Cutout(
        path="sarah_first",
        module=["sarah", "era5"],
        time=slice("2013-01-01", "2013-01-01"),
        x=slice(X0, X1),
        y=slice(Y0, Y1),
    )
    modules = cutout.available_features.index.unique("module")
    assert len(modules) == 2
    assert len(cutout.available_features) > len(ref.available_features)


def test_sel(ref):
    cutout = ref.sel(x=slice(X0 + 2, X1 - 1), y=slice(Y0 + 1, Y1 - 2))
    assert cutout.coords["x"][0] - ref.coords["x"][0] == 2
    assert cutout.coords["y"][-1] - ref.coords["y"][-1] == -2


def test_layout_from_area_density(ref):
    density = 0.01  # MW / m^2
    crs = 3035  # in meter
    layout = ref.uniform_density_layout(density, crs=crs)
    assert layout.dims == ("y", "x")
    assert layout.shape == (ref.data.y.size, ref.data.x.size)
    assert np.isclose(ref.area(crs=crs).sum() * density, layout.sum())



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Functions for aggregating results.
"""

import dask
import xarray as xr


def aggregate_matrix(da, matrix, index):
    if index.name is None:
        index = index.rename("dim_0")
    if isinstance(da.data, dask.array.core.Array):
        da = da.stack(spatial=("y", "x"))
        da = da.chunk(dict(spatial=-1))
        return xr.apply_ufunc(
            lambda da: da * matrix.T,
            da,
            input_core_dims=[["spatial"]],
            output_core_dims=[[index.name]],
            dask="parallelized",
            output_dtypes=[da.dtype],
            dask_gufunc_kwargs=dict(output_sizes={index.name: index.size}),
        ).assign_coords(**{index.name: index})
    else:
        da = da.stack(spatial=("y", "x")).transpose("spatial", "time")
        return xr.DataArray(matrix * da, [index, da.coords["time"]])



# SPDX-FileCopyrightText: Contributors to atlite <https://github.com/pypsa/atlite>
#
# SPDX-License-Identifier: MIT
"""
Module involving hydro operations in atlite.
"""

import logging
from collections import namedtuple

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from shapely.geometry import Point
from tqdm import tqdm

logger = logging.getLogger(__name__)

Basins = namedtuple("Basins", ["plants", "meta", "shapes"])


def find_basin(shapes, lon, lat):
    hids = shapes.index[shapes.intersects(Point(lon, lat))]
    if len(hids) > 1:
        logger.warning(
            f"The point ({lon}, {lat}) is in several basins: {hids}. "
            "Assuming the first one."
        )
    return hids[0]


def find_upstream_basins(meta, hid):
    hids = [hid]
    i = 0
    while i < len(hids):
        hids.extend(meta.index[meta["NEXT_DOWN"] == hids[i]])
        i += 1
    return hids


def determine_basins(plants, hydrobasins, show_progress=False):
    if isinstance(hydrobasins, str):
        hydrobasins = gpd.read_file(hydrobasins)

    assert isinstance(hydrobasins, gpd.GeoDataFrame), (
        "hydrobasins should be passed as a filename or a GeoDataFrame, "
        f"but received `type(hydrobasins) = {type(hydrobasins)}`"
    )

    missing_columns = pd.Index(
        ["HYBAS_ID", "DIST_MAIN", "NEXT_DOWN", "geometry"]
    ).difference(hydrobasins.columns)
    assert (
        missing_columns.empty
    ), "Couldn't find the column(s) {} in the hydrobasins dataset.".format(
        ", ".join(missing_columns)
    )

    hydrobasins = hydrobasins.set_index("HYBAS_ID")

    meta = hydrobasins[hydrobasins.columns.difference(("geometry",))]
    shapes = hydrobasins["geometry"]

    plant_basins = []
    for p in tqdm(
        plants.itertuples(),
        disable=not show_progress,
        desc="Determine upstream basins per plant",
    ):
        hid = find_basin(shapes, p.lon, p.lat)
        plant_basins.append((hid, find_upstream_basins(meta, hid)))
    plant_basins = pd.DataFrame(
        plant_basins, columns=["hid", "upstream"], index=plants.index
    )

    unique_basins = pd.Index(plant_basins["upstream"].sum()).unique().rename("hid")
    return Basins(plant_basins, meta.loc[unique_basins], shapes.loc[unique_basins])


def shift_and_aggregate_runoff_for_plants(
    basins, runoff, flowspeed=1, show_progress=False
):
    inflow = xr.DataArray(
        np.zeros((len(basins.plants), runoff.indexes["time"].size)),
        [("plant", basins.plants.index), runoff.coords["time"]],
    )

    for ppl in tqdm(
        basins.plants.itertuples(),
        disable=not show_progress,
        desc="Shift and aggregate runoff by plant",
    ):
        inflow_plant = inflow.loc[dict(plant=ppl.Index)]
        distances = (
            basins.meta.loc[ppl.upstream, "DIST_MAIN"]
            - basins.meta.at[ppl.hid, "DIST_MAIN"]
        )
        nhours = (distances / (flowspeed * 3.6) + 0.5).astype(int)

        for b in ppl.upstream:
            inflow_plant += runoff.sel(hid=b).roll(time=nhours.at[b])

    return inflow
